#!/usr/bin/env python
""" """

# Standard library modules.
import unittest
import logging
import io

# Third party modules.

# Local modules.
from pypenelopetools.pengeom.module import Module
from pypenelopetools.pengeom.surface import SurfaceImplicit
from pypenelopetools.material import Material, VACUUM

# Globals and constants variables.

class TestModule(unittest.TestCase):

    LINES1 = ['MODULE  (   1) Test',
            'MATERIAL(   1)',
            'SURFACE (   1), SIDE POINTER=(-1)',
            'SURFACE (   2), SIDE POINTER=( 1)',
            'MODULE  (   2)',
            '1111111111111111111111111111111111111111111111111111111111111111',
            '  OMEGA=(+0.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
            '  THETA=(+0.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
            '    PHI=(+1.800000000000000E+02,   0) DEG          (DEFAULT=0.0)',
            'X-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
            'Y-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
            'Z-SHIFT=(-1.000000000000000E+05,   0)              (DEFAULT=0.0)',
            '0000000000000000000000000000000000000000000000000000000000000000']
    LINES2 = ['MODULE  (   2) ',
            'MATERIAL(   0)',
            '1111111111111111111111111111111111111111111111111111111111111111',
            '  OMEGA=(+0.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
            '  THETA=(+0.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
            '    PHI=(+0.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
            'X-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
            'Y-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
            'Z-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
            '0000000000000000000000000000000000000000000000000000000000000000']

    def setUp(self):
        super().setUp()

        self.mat1 = Material('copper', {29: 1.0}, 8.9)

        self.surface1 = SurfaceImplicit()
        self.surface2 = SurfaceImplicit()

        self.module2 = Module(VACUUM)

        self.module1 = Module(self.mat1, 'Test')
        self.module1.add_surface(self.surface1, -1)
        self.module1.add_surface(self.surface2, 1)
        self.module1.add_module(self.module2)
        self.module1.rotation.phi_deg = 180
        self.module1.shift.z_cm = -1e5

    def _test_module1(self, module1):
        self.assertEqual('copper', module1.material.name)
        self.assertEqual('Test', module1.description)
        self.assertAlmostEqual(180, module1.rotation.phi_deg, 4)
        self.assertAlmostEqual(-1e5, module1.shift.z_cm, 4)
        self.assertEqual(2, len(module1.get_surfaces()))
        self.assertEqual(1, len(module1.get_modules()))

    def _test_module2(self, module2):
        self.assertEqual(str(VACUUM), str(module2.material))
        self.assertEqual(0, len(module2.get_surfaces()))
        self.assertEqual(0, len(module2.get_modules()))

    def testskeleton(self):
        self._test_module1(self.module1)
        self._test_module2(self.module2)

    def test_write_read(self):
        material_lookup = {0: VACUUM, 1: self.mat1}
        surface_lookup = {1: self.surface1, 2: self.surface2}
        module_lookup = {2: self.module2}
        index_lookup = {VACUUM: 0, self.mat1: 1,
                        self.surface1: 1, self.surface2: 2,
                        self.module1: 1, self.module2: 2}

        # Module 1
        fileobj = io.StringIO()

        try:
            self.module1._write(fileobj, index_lookup)

            lines = fileobj.getvalue().splitlines()
            self.assertListEqual(self.LINES1, lines)

            fileobj.seek(0)
            module = Module()
            module._read(fileobj, material_lookup, surface_lookup, module_lookup)
            self._test_module1(module)
        finally:
            fileobj.close()

        # Module 2
        fileobj = io.StringIO()

        try:
            self.module2._write(fileobj, index_lookup)

            lines = fileobj.getvalue().splitlines()
            self.assertListEqual(self.LINES2, lines)

            fileobj.seek(0)
            module = Module()
            module._read(fileobj, material_lookup, surface_lookup, module_lookup)
            self._test_module2(module)
        finally:
            fileobj.close()

#    def test_create_lines(self):
#        # Module 1
#        lines = self.module1._create_lines(self.index_lookup)
#        self.assertEqual(12, len(lines))
#        self.assertEqual(self.LINES1, lines)
#
#        # Module 2
#        lines = self.module2._create_lines(self.index_lookup)
#        self.assertEqual(9, len(lines))
#        self.assertEqual(self.LINES2, lines)
#
#    def test_parse_lines(self):

#
#        module1 = Module._parse_lines(self.LINES1, material_lookup, surface_lookup, module_lookup)
#        self._test_module1(module1)
#
#        module2 = Module._parse_lines(self.LINES2, material_lookup, surface_lookup, module_lookup)
#        self._test_module2(module2)

if __name__ == '__main__': #pragma: no cover
    logging.getLogger().setLevel(logging.DEBUG)
    unittest.main()



""""""

# Standard library modules.
import unittest
import logging
import os
import io

# Third party modules.

# Local modules.
from pypenelopetools.material import Material

# Globals and constants variables.

class TestMaterial(unittest.TestCase):

    def setUp(self):
        super().setUp()

        self.mat = Material('mat1', {29: 0.4, 30: 0.6}, 8.9, 326.787, 2.686, 13.496)

        self.testdatadir = os.path.join(os.path.dirname(__file__), 'testdata')

    def _test_material(self, mat):
        self.assertEqual('mat1', mat.name)
        self.assertAlmostEqual(0.4, mat.composition[29], 4)
        self.assertAlmostEqual(0.6, mat.composition[30], 4)
        self.assertAlmostEqual(8.9, mat.density_g_per_cm3, 4)
        self.assertAlmostEqual(326.787, mat.mean_excitation_energy_eV, 4)
        self.assertAlmostEqual(2.686, mat.oscillator_strength_fcb, 4)
        self.assertAlmostEqual(13.496, mat.plasmon_energy_wcb_eV, 4)

    def test__init__(self):
        self._test_material(self.mat)

    def testwriteread_input(self):
        fileobj = io.StringIO()

        try:
            self.mat.write_input(fileobj)

            fileobj.seek(0)
            mat = Material.read_input(fileobj)

            self._test_material(mat)
        finally:
            fileobj.close()

    def testread_material(self):
        filepath = os.path.join(self.testdatadir, 'material', 'mat1.mat')
        with open(filepath, 'r') as fp:
            mat = Material.read_material(fp)

        self._test_material(mat)

if __name__ == '__main__': #pragma: no cover
    logging.basicConfig()
    logging.getLogger().setLevel(logging.DEBUG)
    unittest.main()



"""
Separators used specifically for PENEPMA.
"""

# Standard library modules.

# Third party modules.

# Local modules.
from pypenelopetools.penelope.separator import Separator

# Globals and constants variables.

GEOMETRY = Separator('>>>>>>>> Geometry of the sample.')
"""Section for geometry."""

PHOTON_DETECTORS = Separator('>>>>>>>> Photon detectors (up to 25 different detectors).')
"""Section for photon detector(s)."""

SPATIAL_DISTRIBUTION = Separator('>>>>>>>> Spatial distribution of events in a box.')
"""Section for spatial distribution."""



"""
Geometry definition for PENGEOM.
"""

# Standard library modules.
import os
from itertools import chain
from operator import methodcaller, attrgetter

# Third party modules.

# Local modules.
from pypenelopetools.pengeom.mixin import ModuleMixin
from pypenelopetools.pengeom.module import Module
from pypenelopetools.pengeom.surface import SurfaceImplicit, SurfaceReduced
from pypenelopetools.pengeom.base import _GeometryBase, LINE_SIZE, LINE_START, LINE_SEPARATOR, LINE_END
from pypenelopetools.material import VACUUM

# Globals and constants variables.

def _topological_sort(d, k):
    """
    Togological sort.
    http://stackoverflow.com/questions/108586/topological-sort-recursive-using-generators
    """
    for ii in d.get(k, []):
        yield from _topological_sort(d, ii)
    yield k

class Geometry(ModuleMixin, _GeometryBase):
    """
    Creates a new PENELOPE geometry.
    
    Args:
        title (str, optional): 
            Title of the geometry.
        tilt_deg (float, optional): 
            Specimen tilt in degrees along the x-axis.
        rotation_deg (float, optional): 
            Specimen rotation in degrees along the z-axis
    """

    def __init__(self, title="Untitled", tilt_deg=0.0, rotation_deg=0.0):
        self.title = title
        self.tilt_deg = tilt_deg
        self.rotation_deg = rotation_deg
        self._modules = set()

    def _read(self, fileobj, material_lookup, surface_lookup, module_lookup):
        line = self._read_next_line(fileobj)
        if line != LINE_START:
            raise IOError('Expected start line')

        line = self._read_next_line(fileobj)
        self.title = ''
        while line != LINE_SEPARATOR:
            line = line.lstrip('C').strip()
            self.title += line
            line = self._read_next_line(fileobj)

        line = self._peek_next_line(fileobj)
        while line != LINE_END:
            # Parse section name
            section_name, index, _ = self._parse_line(line)
            index = int(index)

            # Parse surface or module
            if section_name == 'SURFACE':
                # Read 2 lines down the INDICES line
                offset = fileobj.tell()
                self._read_next_line(fileobj)
                indices_line = self._read_next_line(fileobj)
                _, indices, _ = self._parse_line(indices_line)
                indices = map(int, indices.split(','))
                fileobj.seek(offset)

                if sum(indices) == 0:
                    surface = SurfaceImplicit()
                else:
                    surface = SurfaceReduced()

                surface._read(fileobj, material_lookup, surface_lookup, module_lookup)
                surface_lookup[index] = surface

            elif section_name == 'MODULE':
                module = Module()
                module._read(fileobj, material_lookup, surface_lookup, module_lookup)
                self.add_module(module)
                module_lookup[index] = module

            else:
                raise IOError('Cannot read {} section'.format(section_name))

            # Next line
            line = self._peek_next_line(fileobj).rstrip()

    def read(self, fileobj, material_lookup):
        """
        Reads a geometry file (``.geo``).
        
        Args:
            fileobj (file object): 
                File object opened with read access.
            material_lookup (dict(int, :class:`Material <pypenelopetools.material.Material>`)): 
                A lookup table for the materials used in the geometry. 
                Dictionary where the keys are material indexes in the geometry 
                file and the values, 
                :class:`Material <pypenelopetools.material.Material>` instances.
        """
        surface_lookup = {}
        module_lookup = {}
        material_lookup.setdefault(0, VACUUM)

        self._read(fileobj, material_lookup, surface_lookup, module_lookup)

    def _write(self, fileobj, index_lookup):
        fileobj.write(LINE_START + os.linesep)
        fileobj.write('       ' + self.title + os.linesep)
        fileobj.write(LINE_SEPARATOR + os.linesep)

        # Surfaces
        surfaces = sorted((index_lookup[surface], surface)
                          for surface in self.get_surfaces())

        for _index, surface in surfaces:
            surface._write(fileobj, index_lookup)

        # Modules
        modules = sorted((index_lookup[module], module)
                          for module in self.get_modules())

        for _index, module in modules:
            module._write(fileobj, index_lookup)

        # Extra module for tilt and rotation
        if self.tilt_deg != 0.0 or self.rotation_deg != 0.0:
            extra = self._create_extra_module()

            index_lookup[extra] = len(self.get_modules()) + 1
            extra._write(fileobj, index_lookup)

        # End of line
        fileobj.write(LINE_END + os.linesep)

    def write(self, fileobj, index_lookup=None):
        """
        Writes the geometry file (``.geo``) to create this geometry.
        
        Args:
            fileobj (file object): 
                File object opened with write access.
            index_lookup (dict(:obj:`_GeometryBase <pypenelopetools.pengeom.base._GeometryBase>`, int), optional): 
                A lookup table for the surfaces, modules and materials of this 
                geometry. 
                If ``None``, the index lookup is generated by the method 
                :meth:`indexify <pypenelopetools.pengeom.geometry.Geometry.indexify>`.
            
        Returns:
            dict(:obj:`_GeometryBase <pypenelopetools.pengeom.base._GeometryBase>`, int): lookup table
        """
        if not index_lookup:
            index_lookup = self.indexify()
        self._write(fileobj, index_lookup)
        return index_lookup

    def get_materials(self):
        """
        Returns all materials in this geometry.
        """
        return set(map(attrgetter('material'), self.get_modules()))

    def get_surfaces(self):
        """
        Returns all surfaces in this geometry.
        """
        return set(chain(*map(methodcaller('get_surfaces'), self.get_modules())))

    def indexify(self):
        """
        Returns a lookup table which associates the surfaces, modules and 
        materials of this geometry to their index used in the geometry file. 
        The lookup table is a dictionary where the keys are surfaces, modules 
        and materials instances, and the values, an integer index.
        """
        index_lookup = {}

        # Materials
        index_lookup[VACUUM] = 0
        for i, material in enumerate(self.get_materials(), 1):
            index_lookup[material] = i

        # Surfaces
        for i, surface in enumerate(self.get_surfaces(), 1):
            index_lookup[surface] = i

        # Modules
        modules_dep = {} # module dependencies
        for module in self.get_modules():
            modules_dep.setdefault(module, [])
            for submodule in module.get_modules():
                modules_dep[module].append(submodule)

        modules_order = []
        for module in modules_dep:
            for dep_module in _topological_sort(modules_dep, module):
                if dep_module not in modules_order:
                    modules_order.append(dep_module)

        for i, module in enumerate(modules_order, 1):
            index_lookup[module] = i

        return index_lookup

    def _create_extra_module(self):
        extra = Module(VACUUM, description='Extra module for rotation and tilt')

        ## Find all unlinked modules
        all_modules = set(self.get_modules())
        linked_modules = set(chain(*map(methodcaller('get_modules'), all_modules)))
        unlinked_modules = all_modules - linked_modules
        for module in unlinked_modules:
            extra.add_module(module)

        ## Change of Euler angles convention from ZXZ to ZYZ
        extra.rotation.omega_deg = (self.rotation_deg - 90.0) % 360.0
        extra.rotation.theta_deg = self.tilt_deg
        extra.rotation.phi_deg = 90.0

        return extra

    @property
    def title(self):
        """
        Title of the geometry.
        The title must have less than 61 characters.
        """
        return self._title

    @title.setter
    def title(self, title):
        if len(title) > LINE_SIZE - 3:
            raise ValueError("The length of the title ({0:d}) must be less than {1:d}."
                             .format(len(title), LINE_SIZE - 3))
        self._title = title

    @property
    def tilt_deg(self):
        """
        Specimen tilt in degrees along the x-axis
        """
        return self._tilt_deg

    @tilt_deg.setter
    def tilt_deg(self, angle_deg):
        while angle_deg < 0:
            angle_deg += 360.0
        self._tilt_deg = angle_deg

    @property
    def rotation_deg(self):
        """
        Specimen rotation in degrees along the z-axis
        """
        return self._rotation_deg

    @rotation_deg.setter
    def rotation_deg(self, angle_deg):
        self._rotation_deg = angle_deg



"""
Definition of base separator classes.
"""

# Standard library modules.
import os

# Third party modules.

# Local modules.
from pypenelopetools.penelope.base import _InputLineBase

# Globals and constants variables.

class Separator(_InputLineBase):
    """
    Base of all PENELOPE separators.
    
    Args:
        text (str): Comment of the separator
        name (str, optional): name of separator keyword (e.g. END)
        
    Attributes:
        text (str): Comment of the separator
        name (str, optional): name of separator keyword (e.g. END)
    """

    def __init__(self, text, name=''):
        self.text = text
        self.name = name

    def __str__(self):
        if self.name:
            return '{0} {1}'.format(self.name, self.text)
        else:
            return self.text

    def read(self, fileobj):
        if not self.name:
            return

        line = self._peek_next_line(fileobj)
        name = line[:6].strip()

        if name != self.name:
            return

        self._read_next_line(fileobj)

    def write(self, fileobj):
        line = self._create_line(self.name, (self.text,))
        fileobj.write(line + os.linesep)




""""""

# Standard library modules.

# Third party modules.

# Local modules.
from pypenelopetools.penelope.keyword import \
    TypeKeyword, KeywordSequence, KeywordGroup
import pypenelopetools.penelope.keywords as penelope_keywords

# Globals and constants variables.

class SBOX(TypeKeyword):
    """
    Extended source box. The source has uniform activity within
    the volume of a right prism centred at the point (SX0,SY0,
    SZ0) and whose sides have lengths SSX, SSY and SSZ.
      DEFAULT: SSX=SSY=SSZ=0.0
    """

    def __init__(self):
        super().__init__("SPOSIT", (float, float, float),
                         comment="Source box dimensions")

    def set(self, ssx, ssy, ssz):
        super().set(ssx, ssy, ssz)

class SBODY(KeywordSequence):
    """
    In the case of a extended source, the active volume can be restricted
    to that of a body or a set of bodies, which must be defined as parts
    of the geometry. The activity of the source is assumed to be uniform
    within the volume of the intersection of the active bodies and the
    source box. Note that the initial coordinates of primary particles
    are sampled by the rejection method; the sampling efficiency is equal
    to the fraction of the source box volume that is occupied by active
    bodies.
    
    To define each active source body, add the following keyword.
    
    Active source body (PENGEOM sequential body label). 
    One line for each body.
    DEFAULT: None
    
    The program stops if the source box has not been defined previously.
    """

    def __init__(self, maxlength=5000):
        keyword = TypeKeyword("SBODY", (int,),
                              comment='Active source body; one line for each body')
        super().__init__(keyword, maxlength)

    def add(self, kb):
        """
        :arg kb: index of the body or a 
            :class:`Module <pypenelopetools.pengeom.module.Module>` object from 
            a :class:`Geometry <penelopetools.pengeom.geometry.Geometry>` object
        """
        return super().add(kb)

class IPSFN(KeywordSequence):
    """
    The initial state variables of primary particles can be read directly
    from a set of pre-calculated phase-space files (psf). When this
    option is active, previous definitions about the source are ignored.
    Photons from the psf's are assumed to be unpolarised.
    
    Name of an input psf (up to 20 characters).
        DEFAULT: none
    Up to 100 psf's may be declared. They are read sequentially.
    
    The input psf is in ASCII format. Each line defines the initial state
    of a particle; it contains the following quantities in free format
    (and in the order they are listed here):
    -- KPAR, type of particle (1, electron; 2, photon; 3, positron).
    -- E, energy.
    -- X,Y,Z, position coordinates.
    -- U,V,W, direction cosines.
    -- WGHT, weight.
    -- ILB(1),ILB(2),ILB(3),ILB(4), a set of indices that provide
           information on how the particle was generated (see the file
           'manual.txt').
    -- NSHI, incremental shower number (difference between the shower
           numbers of the present particle and the one preceding it
           in the psf).
    Phase-space files can be generated by running PENMAIN using an impact
    detector with the flag IPSF=1 (see below).
    """

    def __init__(self, maxlength=100):
        keyword = TypeKeyword("IPSFN", (str,),
                              comment="Input psf name, up to 20 characters")
        super().__init__(keyword, maxlength)

    def add(self, filename):
        return super().add(filename)

class IPSPLI(TypeKeyword):
    """
    Because of the limited size of the psf's, the results of analogue
    simulations tend to be 'too noisy'. This can be partially corrected
    by splitting the particles from the psf.
    
    Splitting number. Each particle in the psf's will be split
    into NSPLIT equivalent particles, with weights equal to
    WGHT/NSPLIT.
      DEFAULT: NSPLIT=1 (no splitting)
    """

    def __init__(self):
        super().__init__("IPSPLI", (int,),
                         comment="Splitting number")

    def set(self, nsplit):
        super().set(nsplit)

class WGTWIN(TypeKeyword):
    """
    Weight window, (WGMIN,WGMAX). Particles in the phase-space
    file that have initial weights WGHT less than WGMIN will be
    subjected to Russian roulette, and those with WGHT larger
    than WGMAX will be split. Note that the weight window has
    preference over the splitting option, i.e., a particle will
    be split into NSPLIT or less particles only if the latter
    have weights larger than WGMIN.
      DEFAULTS: WGMIN=1.0E-35, WGMAX=1.0E35  (no action)
    """

    def __init__(self):
        super().__init__("WGTWIN", (float, float),
                         comment="Weight window, RR & spl of psf particles")

    def set(self, wgmin, wgmax):
        super().set(wgmin, wgmax)

class EPMAX(TypeKeyword):
    """
    Maximum energy (in eV) of particles in the psf's.
    EPMAX is the upper limit of the energy interval covered by
    the simulation lookup tables. To minimize interpolation
    errors, EPMAX should not be much larger than the maximum
    energy actually occurring during the simulation.
    
    When the initial state variables of particles are read from
    a psf, this parameter is required to initialise PENELOPE and
    is critical; the code crashes if it finds a particle that
    has energy larger than EPMAX.
      DEFAULT: EPMAX=1.0E9 (interpolation is not optimal)
    """

    def __init__(self):
        super().__init__("EPMAX", (float,),
                         comment="Maximum energy of particles in the psf")

    def set(self, epmax):
        super().set(epmax)

class PARINP(KeywordSequence):
    """
    The values of certain parameters of the geometry definition
    may be defined from the main program by means of the array
    PARINP (an input argument of the GEOMIN subroutine). The
    entered PARINP(IP) value replaces the parameter values that
    are marked with the index IP in the geometry definition
    file.
      DEFAULT:  none
    """

    def __init__(self, maxlength=500):
        keyword = TypeKeyword("PARINP", (int, float),
                              comment="Replacement parameter")
        super().__init__(keyword, maxlength)

    def add(self, ip, parinp):
        return super().add(ip, parinp)

class ImpactDetectorGroup(KeywordGroup):
    """
    Each impact detector consists of a set of active bodies, which must
    have been defined as parts of the geometry. The output spectrum is
    the energy distribution of particles that entered any of the active
    bodies coming from a body that is not active (i.e. that is not part
    of the detector). Notice that a detected particle can re-enter the
    detector volume and, consequently, be 'counted' several times (except
    when the flag IDCUT is set equal to 0, see below).
    
    Active bodies cannot be void, because the geometry routines would not
    stop particles at their limiting surfaces. In case you need to define
    detectors outside the material system, fill them with an arbitrary
    material of very small density to avoid perturbing the transport
    process.
    
    To define each impact detector, insert the following block of lines;
    
    IMPDET : Starts the definition of a new detector. Up to 25 different
             detectors can be considered.
             EL and EU are the lower and upper limits of the energy
               window covered by the impact detector.
             NBE is the number of bins in the output energy spectrum of
               the detector (.LE. 1000). If NBE is positive, energy bins
               have uniform width, DE=(EU-EL)/NBE. When NBE is negative,
               the bin width increases geometrically with the energy,
               i.e., the energy bins have uniform width on a logarithmic
               scale.
    
             The integer flag IPSF serves to activate the creation of a
             phase-space file (psf), which contains the state variables
             of all particles that enter the detector. Use this option
             with care, because psf's may grow very fast.
             IPSF=0; no psf is created.
             IPSF=1; the psf is created. Only one PSF can be created in
               each simulation run.
    
             The integer flag IDCUT allows discontinuing the tracking of
             particles that enter the detector.
             IDCUT=0; the simulation of a particle is discontinued when
               it enters the detector (useful to stop the simulation of
               particles recorded in a psf).
             IDCUT=1; the presence of the detector does not affect the
               tracking of particles.
             IDCUT=2; the presence of the detector does not affect the
               tracking of particles. The distribution of particle
               fluence with respect to energy (integrated over the volume
               of the detector) is tallied. The calculated distribution
               has dimensions of length/energy.
    
               DEFAULTS: None
    
    IDPSF_ : Name of the output phase-space file (up to 20 characters).
               DEFAULT: 'psf-impdet-##.dat'
    
    IDSPC_ : Name of the output energy spectrum file (up to 20
             characters).
               DEFAULT: 'spc-impdet-##.dat'
    
    IDFLNC : Name of the output file with the energy distribution of
             particle fluence (20 characters). This file is generated
             only when IDCUT=2.
               DEFAULT: 'fln-impdet-##.dat'
    
    IDAGEL : Activates the evaluation of the age of particles, defined as
             the time elapsed since the start of the primary particle
             that originated the shower. The program generates the age
             distribution of detected particles, i.e., particles of the
             types declared in lines IDKPAR (see below) that enter the
             detector with energy in the window (EL,EU). The distribution
             is tallied for ages in the interval between AGEL and AGEU
             (both in seconds), which is partitioned into NAGE bins. If
             NAGE is positive, the age bins have uniform width. When
             NAGE is negative, the width of age bins is uniform on a
             logarithmic scale.
    
               DEFAULTS: NAGE=100, AGEL=0.0, AGEU must always be
                         specified
    
    IDAGEF : Name of the output age distribution file (up to 20
             characters)
               DEFAULT: 'age-impdet-##.dat'
    
    IDBODY : Active body of the detector. One line for each active body.
               DEFAULT: None
             --> Notice that a body cannot be part of more than one
             impact detector.
    
    IDKPAR : Type of particle that is detected (1=electrons, 2=photons or
             3=positrons). One line for each type.
    
             The detector has no effect for particles that are not
             detected. This feature can be used, e.g., to make a body or
             a set of bodies opaque to particles of a certain type.
    
               DEFAULT: All particles are detected
   """

    def __init__(self):
        self.IMPDET = TypeKeyword('IMPDET', (float, float, int, int, int),
                                  comment='E-window, no. of bins, IPSF, IDCUT')
        self.IDSPC = TypeKeyword('IDSPC', (str,),
                                 comment='Spectrum file name, 20 chars')
        self.IDPSF = TypeKeyword('IDPSF', (str,),
                                 comment='Phase-space file name, 20 chars')
        self.IDFLNC = TypeKeyword('IDFLNC', (str,),
                                  comment='Fluence spectrum file name, 20 chars')
        self.IDAGEL = TypeKeyword('IDAGEL', (float, float, int),
                                  comment='Age interval and no. of bins')
        self.IDAGEF = TypeKeyword('IDAGEF', (str,),
                                  comment='Age-distribution file name, 20 chars')

        keyword = TypeKeyword('IDBODY', (int,), comment='Active body')
        self.IDBODY = KeywordSequence(keyword, maxlength=5000)

        keyword = TypeKeyword('IDKPAR', (int,), comment='Kind of detected particles')
        self.IDKPAR = KeywordSequence(keyword, maxlength=3)

    def get_keywords(self):
        return (self.IMPDET, self.IDSPC, self.IDPSF, self.IDFLNC,
                self.IDAGEL, self.IDAGEF, self.IDBODY, self.IDKPAR)

    def set(self, el, eu, nbe, ipsf, idcut,
            spectrum_filename=None, psf_filename=None, fln_filename=None,
            agel=None, ageu=None, nage=None, age_filename=None,
            kb=None, kpar=None):
        self.IMPDET.set(el, eu, nbe, ipsf, idcut)
        self.IDSPC.set(spectrum_filename)
        self.IDPSF.set(psf_filename)
        self.IDFLNC.set(fln_filename)
        self.IDAGEL.set(agel, ageu, nage)
        self.IDAGEF.set(age_filename)
        self._set_keyword_sequence(self.IDBODY, kb)
        self._set_keyword_sequence(self.IDKPAR, kpar)

class ImpactDetectors(KeywordSequence):

    def __init__(self, maxlength=25):
        keyword = ImpactDetectorGroup()
        super().__init__(keyword, maxlength)

    def add(self, el, eu, nbe, ipsf, idcut,
            spectrum_filename=None, psf_filename=None, fln_filename=None,
            agel=None, ageu=None, nage=None, age_filename=None,
            kb=None, kpar=None):
        return super().add(el, eu, nbe, ipsf, idcut,
                           spectrum_filename, psf_filename, fln_filename,
                           agel, ageu, nage, age_filename,
                           kb, kpar)

class EnergyDepositionDetectorGroup(KeywordGroup):
    """
    Each energy-deposition detector consists of a set of active bodies,
    which must have been defined as parts of the geometry. The output
    spectrum is the distribution of absorbed energy (per primary shower)
    in the active bodies.
    
    .. warning::
       The energy-deposition spectrum may be strongly biased when interaction 
       forcing is applied, even outside the detector bodies.
    """

    def __init__(self):
        self.ENDETC = penelope_keywords.ENDETC()
        self.EDSPC = penelope_keywords.EDSPC()

        keyword = TypeKeyword('EDBODY', (int,), comment='Active body')
        self.EDBODY = KeywordSequence(keyword, maxlength=5000)

    def get_keywords(self):
        return (self.ENDETC, self.EDSPC, self.EDBODY)

    def set(self, el, eu, nbe, spectrum_filename=None, kb=None):
        self.ENDETC.set(el, eu, nbe)
        self.EDSPC.set(spectrum_filename)
        self._set_keyword_sequence(self.EDBODY, kb)

class EnergyDepositionDetectors(KeywordSequence):

    def __init__(self, maxlength=25):
        keyword = EnergyDepositionDetectorGroup()
        super().__init__(keyword, maxlength)

    def add(self, el, eu, nbe, spectrum_filename=None, kb=None):
        return super().add(el, eu, nbe, spectrum_filename, kb)

class GRIDR(TypeKeyword):
    """
    The efficiency of the dose map calculation can be increased by taking
    advantage of possible symmetries of the system (source and geometry).
    In problems with axial symmetry about the Z axis, it is advantageous
    to tally the dose distribution in the volume of a cylinder of radius
    RU, about the Z axis, limited by the planes Z=ZL and Z=ZU. For
    systems with spherical symmetry about the origin of coordinates, it
    is most convenient to consider the radial dose distribution in a
    sphere of radius RU. The generation of these symmetric dose maps is
    activated by entering the line
    
    GRIDR_ : Radius RU of the dose zone and number of radial bins.
               DEFAULT: None
    
    Specifically, when the input file contains only the lines GRIDZ_ and
    GRIDR_, the program assumes that the dose distribution is axially
    symmetric and generates a cylindrical map. When the input file has
    only the line GRIDR_, spherical symmetry is assumed and the radial
    distribution of absorbed dose is tallied.
    
    The different types of dose maps are mutually exclusive. Notice that
    when the assumed symmetry does not hold, the program may not be able
    to evaluate the masses of voxels correctly.
    """

    def __init__(self):
        super().__init__('GRIDR', (float, int),
                         comment='Radius of the dose volume, no. of bins')

    def set(self, ru, ndbr):
        super().set(ru, ndbr)







""""""

# Standard library modules.

# Third party modules.

# Local modules.
from pypenelopetools.penelope.separator import Separator

# Globals and constants variables.

INPUT_PHASE_SPACE_FILE = Separator('>>>>>>>> Input phase-space file (psf).')
GEOMETRY = Separator('>>>>>>>> Geometry and local simulation parameters.')
IMPACT_DETECTORS = Separator('>>>>>>>> Impact detectors (up to 25 different detectors).')
ABSORBED_DOSE_DISTRIBUTION = Separator('>>>>>>>> Absorbed dose distribution.')



#!/usr/bin/env python
""" """

# Standard library modules.
import unittest
import logging
import os
import io

# Third party modules.

# Local modules.
from pypenelopetools.pengeom.geometry import Geometry
from pypenelopetools.pengeom.surface import zplane, cylinder, xplane
from pypenelopetools.pengeom.module import Module, SidePointer
from pypenelopetools.material import Material, VACUUM

# Globals and constants variables.

class TestGeometry(unittest.TestCase):

    LINES = ['XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX',
               '       Test Geometry',
               '0000000000000000000000000000000000000000000000000000000000000000',
               'SURFACE (   1) Plane Z=0.00 m',
               'INDICES=( 0, 0, 0, 1, 0)',
               'X-SCALE=(+1.000000000000000E+00,   0)              (DEFAULT=1.0)',
               'Y-SCALE=(+1.000000000000000E+00,   0)              (DEFAULT=1.0)',
               'Z-SCALE=(+1.000000000000000E+00,   0)              (DEFAULT=1.0)',
               '  OMEGA=(+0.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
               '  THETA=(+0.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
               '    PHI=(+0.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
               'X-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               'Y-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               'Z-SHIFT=(+1.000000000000000E-08,   0)              (DEFAULT=0.0)',
               '0000000000000000000000000000000000000000000000000000000000000000',
               'SURFACE (   2) Cylinder of radius 0.01 m along z-axis',
               'INDICES=( 1, 1, 0, 0,-1)',
               'X-SCALE=(+1.000000000000000E-02,   0)              (DEFAULT=1.0)',
               'Y-SCALE=(+1.000000000000000E-02,   0)              (DEFAULT=1.0)',
               'Z-SCALE=(+1.000000000000000E+00,   0)              (DEFAULT=1.0)',
               '  OMEGA=(+0.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
               '  THETA=(+0.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
               '    PHI=(+0.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
               'X-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               'Y-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               'Z-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               '0000000000000000000000000000000000000000000000000000000000000000',
               'SURFACE (   3) Plane Z=-0.00 m',
               'INDICES=( 0, 0, 0, 1, 0)',
               'X-SCALE=(+1.000000000000000E+00,   0)              (DEFAULT=1.0)',
               'Y-SCALE=(+1.000000000000000E+00,   0)              (DEFAULT=1.0)',
               'Z-SCALE=(+1.000000000000000E+00,   0)              (DEFAULT=1.0)',
               '  OMEGA=(+0.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
               '  THETA=(+0.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
               '    PHI=(+0.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
               'X-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               'Y-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               'Z-SHIFT=(-1.000000000000000E-01,   0)              (DEFAULT=0.0)',
               '0000000000000000000000000000000000000000000000000000000000000000',
               'SURFACE (   4) Plane X=0.00 m',
               'INDICES=( 0, 0, 0, 1, 0)',
               'X-SCALE=(+1.000000000000000E+00,   0)              (DEFAULT=1.0)',
               'Y-SCALE=(+1.000000000000000E+00,   0)              (DEFAULT=1.0)',
               'Z-SCALE=(+1.000000000000000E+00,   0)              (DEFAULT=1.0)',
               '  OMEGA=(+0.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
               '  THETA=(+9.000000000000000E+01,   0) DEG          (DEFAULT=0.0)',
               '    PHI=(+0.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
               'X-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               'Y-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               'Z-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               '0000000000000000000000000000000000000000000000000000000000000000',
               'MODULE  (   1) ',
               'MATERIAL(   1)',
               'SURFACE (   1), SIDE POINTER=(-1)',
               'SURFACE (   2), SIDE POINTER=(-1)',
               'SURFACE (   3), SIDE POINTER=( 1)',
               'SURFACE (   4), SIDE POINTER=( 1)',
               '1111111111111111111111111111111111111111111111111111111111111111',
               '  OMEGA=(+0.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
               '  THETA=(+0.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
               '    PHI=(+0.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
               'X-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               'Y-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               'Z-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               '0000000000000000000000000000000000000000000000000000000000000000',
               'MODULE  (   2) ',
               'MATERIAL(   2)',
               'SURFACE (   1), SIDE POINTER=(-1)',
               'SURFACE (   2), SIDE POINTER=(-1)',
               'SURFACE (   3), SIDE POINTER=( 1)',
               'MODULE  (   1)',
               '1111111111111111111111111111111111111111111111111111111111111111',
               '  OMEGA=(+0.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
               '  THETA=(+0.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
               '    PHI=(+0.000000000000000E+00,   0) DEG          (DEFAULT=0.0)',
               'X-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               'Y-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               'Z-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               '0000000000000000000000000000000000000000000000000000000000000000',
               'MODULE  (   3) Extra module for rotation and tilt',
               'MATERIAL(   0)',
               'MODULE  (   2)',
               '1111111111111111111111111111111111111111111111111111111111111111',
               '  OMEGA=(+2.700000000000000E+02,   0) DEG          (DEFAULT=0.0)',
               '  THETA=(+4.500000000000000E+01,   0) DEG          (DEFAULT=0.0)',
               '    PHI=(+9.000000000000000E+01,   0) DEG          (DEFAULT=0.0)',
               'X-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               'Y-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               'Z-SHIFT=(+0.000000000000000E+00,   0)              (DEFAULT=0.0)',
               '0000000000000000000000000000000000000000000000000000000000000000',
               'END      0000000000000000000000000000000000000000000000000000000']

    def setUp(self):
        super().setUp()

        self.testdatadir = os.path.join(os.path.dirname(__file__),
                                        '..', 'testdata', 'pengeom')

        self.geo = Geometry('Test Geometry')

        surface1 = zplane(1e-8)
        surface2 = zplane(-1e-1)
        surface3 = cylinder(1.0)
        surface4 = xplane(0.0)

        self.mat1 = Material('copper', {29: 1.0}, 8.9)
        self.module1 = Module(self.mat1)
        self.module1.add_surface(surface1, SidePointer.NEGATIVE)
        self.module1.add_surface(surface2, SidePointer.POSITIVE)
        self.module1.add_surface(surface3, SidePointer.NEGATIVE)
        self.module1.add_surface(surface4, SidePointer.POSITIVE)
        self.geo.add_module(self.module1)

        self.mat2 = Material('zinc', {30: 1.0}, 7.14)
        self.module2 = Module(self.mat2)
        self.module2.add_surface(surface1, SidePointer.NEGATIVE)
        self.module2.add_surface(surface2, SidePointer.POSITIVE)
        self.module2.add_surface(surface3, SidePointer.NEGATIVE)
        self.module2.add_module(self.module1)
        self.geo.add_module(self.module2)

        self.geo.tilt_deg = 45

    def _test_geometry(self, geometry):
        self.assertEqual('Test Geometry', geometry.title)
        self.assertEqual(3, len(geometry.get_modules()))
        self.assertEqual(4, len(geometry.get_surfaces()))
        self.assertEqual(3, len(geometry.get_materials()))

    def testskeleton(self):
        self.assertEqual('Test Geometry', self.geo.title)
        self.assertEqual(2, len(self.geo.get_modules()))
        self.assertEqual(4, len(self.geo.get_surfaces()))
        self.assertEqual(2, len(self.geo.get_materials()))
        self.assertAlmostEqual(45, self.geo.tilt_deg, 4)
        self.assertAlmostEqual(0.0, self.geo.rotation_deg, 4)

    def testindexify(self):
        index_lookup = self.geo.indexify()

        # 4 surfaces, 2 materials, 2 modules, 1 extra module
        self.assertEqual(4 + 2 + 2 + 1, len(index_lookup))

        index_lookup2 = self.geo.indexify()
        self.assertDictEqual(index_lookup, index_lookup2)

    def testwriteread(self):
        material_lookup = {0: VACUUM, 1: self.mat1, 2: self.mat2}
        fileobj = io.StringIO()

        try:
            self.geo.write(fileobj)

            fileobj.seek(0)
            lines = fileobj.getvalue().splitlines()
            self.assertListEqual(self.LINES[:3], lines[:3])
            self.assertEqual(self.LINES[14], lines[14])
            self.assertEqual(self.LINES[26], lines[26])
            self.assertEqual(self.LINES[38], lines[38])
            self.assertEqual(self.LINES[50], lines[50])
            self.assertEqual(self.LINES[51], lines[51])
            self.assertListEqual(self.LINES[57:65], lines[57:65])
            self.assertEqual(self.LINES[65], lines[65])
            self.assertListEqual(self.LINES[71:], lines[71:])

            fileobj.seek(0)
            geometry = Geometry()
            geometry.read(fileobj, material_lookup)

            self._test_geometry(geometry)

            index_lookup = geometry.indexify()
            self.assertEqual(4 + 3 + 3, len(index_lookup))
        finally:
            fileobj.close()

    def test_epma1_read(self):
        material_lookup = {1: self.mat1}

        filepath = os.path.join(self.testdatadir, 'epma1.geo')
        geometry = Geometry()
        with open(filepath, 'r') as fp:
            geometry.read(fp, material_lookup)

        self.assertEqual('Cylindrical homogeneous foil', geometry.title)
        self.assertEqual(3, len(geometry.get_surfaces()))
        self.assertEqual(1, len(geometry.get_modules()))
        self.assertEqual(1, len(geometry.get_materials()))

    def test_epma2_read(self):
        material_lookup = {1: self.mat1, 2: self.mat2}

        filepath = os.path.join(self.testdatadir, 'epma2.geo')
        geometry = Geometry()
        with open(filepath, 'r') as fp:
            geometry.read(fp, material_lookup)

        self.assertEqual('Cylindrical foil with a material couple', geometry.title)
        self.assertEqual(4, len(geometry.get_surfaces()))
        self.assertEqual(2, len(geometry.get_modules()))
        self.assertEqual(2, len(geometry.get_materials()))

if __name__ == '__main__': #pragma: no cover
    logging.getLogger().setLevel(logging.DEBUG)
    unittest.main()



"""
Definition of surfaces.
"""

__all__ = ['SurfaceImplicit', 'SurfaceReduced',
           'xplane', 'yplane', 'zplane', 'cylinder', 'sphere']

# Standard library modules.
import os

# Third party modules.

# Local modules.
from pypenelopetools.pengeom.transformation import Rotation, Shift, Scale
from pypenelopetools.pengeom.base import _GeometryBase, LINE_EXTRA, LINE_SEPARATOR
from pypenelopetools.pengeom.mixin import DescriptionMixin

# Globals and constants variables.

class _SurfaceBase(DescriptionMixin, _GeometryBase):
    """
    Base class for surface definition.
    
    Args:
        description (str): Description of the surface
    """

    def __init__(self, description=''):
        self.description = description

        self._rotation = Rotation()
        self._shift = Shift()

    def _write(self, fileobj, index_lookup):
        index = index_lookup[self]
        text = "{:4d}".format(index)
        comment = " " + self.description
        line = self._create_line("SURFACE", text, comment)
        fileobj.write(line + os.linesep)

    @property
    def rotation(self):
        """:obj:`Rotation <pypenelopetools.pengeom.transformation.Rotation>`: Rotation of the surface."""
        return self._rotation

    @property
    def shift(self):
        """:obj:`Shift <pypenelopetools.pengeom.transformation.Shift>`: Shift/translation of the surface."""
        return self._shift

class SurfaceImplicit(_SurfaceBase):
    """
    Definition of an implicit surface.
    
    Args:
        coefficients (dict(str, float) or list(float)):
            Coefficients for the implicit form of the quadratic equation.
            If the argument is a :obj:`dict`, the keys are the names of 
            coefficient (e.g. ``xx``) and the values the coefficient values.
            If the argument is a :obj:`list`, the list must contain 10 values,
            one for each coefficient.
        description (str): Description of the surface
    """

    def __init__(self, coefficients=None, description=''):
        super().__init__(description)

        if coefficients is None:
            coefficients = [0.0] * 10
        self.coefficients = coefficients

    def __repr__(self):
        coeffs = ['{0}={1}'.format(key, value)
                  for key, value in self.coefficients.items()]
        return '<Surface(description={0}, {1}, rotation={2}, shift={3})>' \
            .format(self.description, ', '.join(coeffs), str(self.rotation), str(self.shift))

    def _read(self, fileobj, material_lookup, surface_lookup, module_lookup):
        line = self._read_next_line(fileobj)
        _, _, self.description = self._parse_line(line)

        line = self._read_next_line(fileobj)
        if line != 'INDICES=( 0, 0, 0, 0, 0)':
            raise IOError('Expected line "INDICES=( 0, 0, 0, 0, 0)" instead of "{0}"'
                          .format(line))

        line = self._read_next_line(fileobj)
        while line != LINE_EXTRA and line != LINE_SEPARATOR:
            keyword, value, _ = self._parse_expline(line)
            key = keyword[1:-1].lower()
            self.coefficients[key] = value
            line = self._read_next_line(fileobj)

        if line == LINE_EXTRA:
            extra_offset = fileobj.tell()
            self.rotation._read(fileobj, material_lookup, surface_lookup, module_lookup)

            fileobj.seek(extra_offset)
            self.shift._read(fileobj, material_lookup, surface_lookup, module_lookup)

    def _create_coefficient_line(self, key):
        value = self.coefficients[key]
        return self._create_expline('A{}='.format(key.upper()), value, '              (DEFAULT=0.0)')

    def _write(self, fileobj, index_lookup):
        super()._write(fileobj, index_lookup)

        # Indices
        text = "{0:2d},{1:2d},{2:2d},{3:2d},{4:2d}".format(0, 0, 0, 0, 0)
        line = self._create_line('INDICES=', text)
        fileobj.write(line + os.linesep)

        # Coefficients
        fileobj.write(self._create_coefficient_line('xx') + os.linesep)
        fileobj.write(self._create_coefficient_line('xy') + os.linesep)
        fileobj.write(self._create_coefficient_line('xz') + os.linesep)
        fileobj.write(self._create_coefficient_line('yy') + os.linesep)
        fileobj.write(self._create_coefficient_line('yz') + os.linesep)
        fileobj.write(self._create_coefficient_line('zz') + os.linesep)
        fileobj.write(self._create_coefficient_line('x') + os.linesep)
        fileobj.write(self._create_coefficient_line('y') + os.linesep)
        fileobj.write(self._create_coefficient_line('z') + os.linesep)
        fileobj.write(self._create_coefficient_line('0') + os.linesep)
        fileobj.write(LINE_EXTRA + os.linesep)

        self.rotation._write(fileobj, index_lookup)
        self.shift._write(fileobj, index_lookup)

        fileobj.write(LINE_SEPARATOR + os.linesep)

    @property
    def coefficients(self):
        """(dict(str, float) or list(float)): Coefficients for the implicit form of the quadratic equation.
        If the value is a :obj:`dict`, the keys are the names of 
        coefficient (e.g. ``xx``) and the values the coefficient values.
        If the argument is a :obj:`list`, the list must contain 10 values,
        one for each coefficient.
        
        Examples:

          >>> s = Surface()
          >>> s.coefficients = {'xx': 0.0, 'xy': 0.0, 'xz': 0.0, 'yy': 0.0, 'yz': 0.0, 'zz': 0.0, 'x': 0.0, 'y': 0.0, 'z': 0.0, '0': 0.0}
          >>> s.coefficients = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
          >>> s.coefficients = {'xx': 1.0, 'xy': 1.0}
        """
        return self._coefficients

    @coefficients.setter
    def coefficients(self, coefficients):
        if isinstance(coefficients, dict):
            self._coefficients = {'xx': 0.0, 'xy': 0.0, 'xz': 0.0,
                                  'yy': 0.0, 'yz': 0.0,
                                  'zz': 0.0,
                                  'x': 0.0, 'y': 0.0, 'z': 0.0, '0': 0.0}

            for key in coefficients:
                key = key.lower()
                assert key in self._coefficients
                self._coefficients[key] = coefficients[key]
        else:
            assert len(coefficients) == 10

            self._coefficients = \
                {'xx': coefficients[0], 'xy': coefficients[1], 'xz': coefficients[2],
                 'yy': coefficients[3], 'yz': coefficients[4],
                 'zz': coefficients[5],
                 'x': coefficients[6], 'y': coefficients[7], 'z': coefficients[8],
                 '0': coefficients[9]}

class SurfaceReduced(_SurfaceBase):
    """
    Definition of a reduced/explicit surface.
    
    Args:
        indices (tuple(int)):
            Indices for the explicit form of the quadratic equation.
            Indices are 5 integers (-1, 0 or 1) defining the surface.
        description (str): Description of the surface
    """

    def __init__(self, indices=None, description=''):
        super().__init__(description)

        if indices is None:
            indices = (0, 0, 0, 0, 0)
        self.indices = indices
        self._scale = Scale()

    def __repr__(self):
        return '<Surface(description={0}, indices={1}, scale={2}, rotation={3}, shift={4})>' \
            .format(self.description, str(self.indices), str(self.scale),
                    str(self.rotation), str(self.shift))

    def _read(self, fileobj, material_lookup, surface_lookup, module_lookup):
        line = self._read_next_line(fileobj)
        _, _, self.description = self._parse_line(line)

        line = self._read_next_line(fileobj)
        keyword, text, _termination = self._parse_line(line)
        if keyword != 'INDICES=':
            raise IOError('Expected keyword "INDICES=" instead of "{0}"'.format(keyword))
        self.indices = tuple(map(int, text.split(',')))

        extra_offset = fileobj.tell()
        self.rotation._read(fileobj, material_lookup, surface_lookup, module_lookup)

        fileobj.seek(extra_offset)
        self.shift._read(fileobj, material_lookup, surface_lookup, module_lookup)

        fileobj.seek(extra_offset)
        self.scale._read(fileobj, material_lookup, surface_lookup, module_lookup)

    def _write(self, fileobj, index_lookup):
        super()._write(fileobj, index_lookup)

        # Indices
        text = "{0:2d},{1:2d},{2:2d},{3:2d},{4:2d}".format(*self.indices)
        line = self._create_line('INDICES=', text)
        fileobj.write(line + os.linesep)

        self.scale._write(fileobj, index_lookup)
        self.rotation._write(fileobj, index_lookup)
        self.shift._write(fileobj, index_lookup)

        fileobj.write(LINE_SEPARATOR + os.linesep)

    @property
    def indices(self):
        """(tuple(int)):Indices for the explicit form of the quadratic equation.
        Indices are 5 integers (-1, 0 or 1) defining the surface.
        """
        return self._indices

    @indices.setter
    def indices(self, indices):
        if len(indices) != 5:
            raise ValueError("Five indices must be defined.")

        for indice in indices:
            if not indice in [-1, 0, 1]:
                raise ValueError("Index ({:d}) must be either -1, 0 or 1.".format(indice))

        self._indices = tuple(indices)

    @property
    def scale(self):
        """:obj:`Scale <pypenelopetools.pengeom.transformation.Scale>`: Scaling of the surface."""
        return self._scale

def xplane(x_cm):
    """
    Returns a surface for a plane X=x.
    
    Args:
        x_cm (float): Intercept on the x-axis (in cm).
    
    Returns:
        :obj:`SurfaceReduced`
    """
    s = SurfaceReduced((0, 0, 0, 1, 0), 'Plane X={:4.2f} cm'.format(x_cm))
    s.shift.x_cm = x_cm
    s.rotation.theta_deg = 90.0
    return s

def yplane(y_cm):
    """
    Returns a surface for a plane Y=y.
    
    Args:
        y_cm (float): Intercept on the y-axis (in cm).

    Returns:
        :obj:`SurfaceReduced`
    """
    s = SurfaceReduced((0, 0, 0, 1, 0), 'Plane Y={:4.2f} cm'.format(y_cm))
    s.shift.y_cm = y_cm
    s.rotation.theta_deg = 90.0
    s.rotation.phi_deg = 90.0
    return s

def zplane(z_cm):
    """
    Returns a surface for a plane Z=z.
    
    Args:
        z_cm (float): Intercept on the z-axis (in cm).

    Returns:
        :obj:`SurfaceReduced`
    """
    s = SurfaceReduced((0, 0, 0, 1, 0), 'Plane Z={:4.2f} cm'.format(z_cm))
    s.shift.z_cm = z_cm
    return s

def cylinder(radius_cm, axis='x'):
    """
    Returns a surface for a cylinder along *axis* with *radius*.

    Args:
        radius_cm (float): Radius of the cylinder (in cm).
        axis (str): Axis of the cylinder, either ``x``, ``y`` or ``z``.
    
    Returns:
        :obj:`SurfaceReduced`
    """
    axis = axis.lower()
    description = 'Cylinder of radius {0:4.2f} cm along {1}-axis'.format(radius_cm, axis)
    s = SurfaceReduced((1, 1, 0, 0, -1), description)

    s.scale.x = radius_cm
    s.scale.y = radius_cm

    if axis == 'z':
        pass
    elif axis == 'x':
        s.rotation.theta_deg = 90.0
    elif axis == 'y':
        s.rotation.theta_deg = 90.0
        s.rotation.phi_deg = 90.0

    return s

def sphere(radius_cm):
    """
    Returns a surface for a sphere or *radius*.
    
    Args:
        radius_cm (float): Radius of the sphere (in cm).

    Returns:
        :obj:`SurfaceReduced`
    """
    description = 'Sphere of radius {:4.2f} cm'.format(radius_cm)
    s = SurfaceReduced((1, 1, 1, 0, -1), description)

    s.scale.x = radius_cm
    s.scale.y = radius_cm
    s.scale.z = radius_cm

    return s



"""
Definition of base result classes.
"""

# Standard library modules.
import abc
import re

# Third party modules.

# Local modules.

# Globals and constants variables.

PATTERN_NUMBER = re.compile(r'\d\.\d*E[\+\-]\d\d')

class _PenelopeResultBase(metaclass=abc.ABCMeta):
    """
    Base class representing a type of result.
    """

    def _read_until_line_startswith(self, fileobj, prefix):
        """
        Reads until a line that starts with *prefix* is found.
        White spaces are ignored at the beginning of each line.
        
        Args:
            fileobj (file object): File object opened with read access.
            prefix (str): Prefix of the line to find.
            
        Returns:
            str: Found line, stripped of all leading and trailing white spaces.
        """
        line = fileobj.readline()
        if not line:
            raise EOFError('Read until EOF, no line with prefix {0}'.format(prefix))

        line = line.strip()
        if line.startswith(prefix):
            return line

        return self._read_until_line_startswith(fileobj, prefix)

    def _read_until_end_of_comments(self, fileobj):
        """
        Read until the end of the comments. 
        The next line returns by the file-object will be a non-comment line.
        
        Args:
            fileobj (file object): File object opened with read access.
        """
        offset = fileobj.tell()
        line = fileobj.readline()
        if not line:
            raise EOFError('Read until EOF')

        line = line.strip()
        if line.startswith('#'):
            return self._read_until_end_of_comments(fileobj)

        fileobj.seek(offset)

    def _read_all_values(self, line):
        """
        Parses all numbers from *line*.
        
        Returns:
            list(float): List of numbers.
        """
        return [float(v) for v in PATTERN_NUMBER.findall(line)]

    @abc.abstractmethod
    def read(self, fileobj):
        """
        Reads a result file.
        
        Args:
            fileobj (file object): File object opened with read access.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def read_directory(self, dirpath):
        """
        Read a result file from a directory.
        
        Args:
            dirpath (str): Path of a directory.
        """
        raise NotImplementedError



"""
Definition of base PENGEOM classes.
"""

# Standard library modules.
import abc
import math
import re

# Third party modules.

# Local modules.
LINE_SIZE = 64
LINE_KEYWORDS_SIZE = 8
LINE_EXTRA = '1' * LINE_SIZE
LINE_START = 'X' * LINE_SIZE
LINE_SEPARATOR = '0' * LINE_SIZE
LINE_END = 'END      0000000000000000000000000000000000000000000000000000000'

PATTERN_LINE = re.compile(r'([A-Z= ]{8})\(([0-9\-\+\, ]*)\)(.*)')
PATTERN_EXPLINE = re.compile(r'([A-Z0-9\-= ]{8})\(([0-9\+\-\.Ee ]*)\,([0-9 ]{4})\)(.*)')

# Globals and constants variables.

def _toexponent(number):
    """
    Formats exponent to PENELOPE format (E22.15)

    Args:
        number (Float): Number to format
    
    Returns:
        str: number to PENELOPE format (E22.15)
    """
    if number == 0.0:
        exponent = 0
    else:
        exponent = int(math.log10(abs(number)))

    if exponent >= 0 :
        exponent_str = '+{:2d}'.format(abs(exponent))
    else:
        exponent_str = '-{:2d}'.format(abs(exponent))

    exponent_str = exponent_str.replace(' ', '0') #Replace white space in the exponent

    coefficient = float(abs(number)) / 10 ** exponent

    if number >= 0:
        number_str = '+{0:17.15f}E{1}'.format(coefficient, exponent_str)
    else:
        number_str = '-{0:17.15f}E{1}'.format(coefficient, exponent_str)

    # The number should not be longer than 22 characters
    assert len(number_str) == 22

    return number_str

class _GeometryBase(metaclass=abc.ABCMeta):
    """
    Base class for geometry objects.
    """

    @abc.abstractmethod
    def _read(self, fileobj, material_lookup, surface_lookup, module_lookup): #@NoSelf
        """
        Reads file object.
        
        Args:
            fileobj (file object): 
                File object opened with read access.
            material_lookup (dict(int, :class:`Material <pypenelopetools.material.Material>`)): 
                A lookup table for the materials used in the geometry. 
                Dictionary where the keys are material indexes in the geometry 
                file and the values, 
                :class:`Material <pypenelopetools.material.Material>` instances.
            surface_lookup (dict(int, :class:`Surface <pypeneloptools.pengeom.surface._SurfaceBase>`)): 
                A lookup table for surfaces used in the geometry. 
                Dictionary where the keys are surface indexes in the geometry 
                file and the values, 
                :class:`Surface <pypenelopetools.pengeom.surface._SurfaceBase>` 
                instances.
            module_lookup (dict(int, :class:`Module <pypenelopetools.pengeom.module.Module>`)): 
                A lookup table for modules used in the geometry. 
                Dictionary where the keys are module indexes in the geometry 
                file and the values, 
                :class:`Module <pypenelopetools.pengeom.module.Module>` instances.
        """
        raise NotImplementedError

    def _peek_next_line(self, fileobj):
        """
        Returns the next line without advancing the current position.
        
        Args:
            fileobj (file object): File object opened with read access.
            
        Returns:
            str: Next line, stripped of all trailing white spaces.
        """
        # Remember the current position
        offset = fileobj.tell()

        # Read line
        line = self._read_next_line(fileobj)

        # Roll back to previous position
        fileobj.seek(offset)

        return line

    def _read_next_line(self, fileobj):
        """
        Returns the next line and advances the current position.
        
        Args:
            fileobj (file object): File object opened with read access.
            
        Returns:
            str: Next line, stripped of all trailing white spaces.
        """
        return fileobj.readline().rstrip()

    def _parse_line(self, line):
        match = PATTERN_LINE.match(line)
        if not match:
            raise IOError('Cannot parse line: "{0}"'.format(line))

        keyword, text, termination = match.groups()
        keyword = keyword.strip()
        text = text.strip()
        termination = termination.strip()

        return keyword, text, termination

    def _parse_expline(self, line):
        match = PATTERN_EXPLINE.match(line)
        if not match:
            raise IOError('Cannot parse line: "{0}"'.format(line))

        keyword, value, index, termination = match.groups()
        keyword = keyword.strip()
        value = float(value)
        index = int(index)
        termination = termination.strip()

        if index != 0:
            raise RuntimeError('Index different than zero is not supported: ' + line)

        return keyword, value, termination

    @abc.abstractmethod
    def _write(self, fileobj, index_lookup): #pragma: no cover
        """
        Writes to file object.
        
        Args:
            fileobj (file object): 
                File object opened with write access.
            index_lookup (dict(:obj:`_GeometryBase <pypenelopetools.pengeom.base._GeometryBase>`, int)): 
                A lookup table for the surfaces, modules and materials of the 
                associated geometry. 
                Each component is assigned an index by the method 
                :meth:`indexify <pypenelopetools.pengeom.geometry.Geometry.indexify>`.
                Dictionary where the keys are surfaces, modules and materials 
                instances, and the values, an integer index.
        """
        raise NotImplementedError

    def _create_line(self, keyword, text, termination=''):
        """
        Creates an input line from the specified keyword, text and comment.
        The white space between the items is automatically adjusted to fit the
        line size.
        The keyword and the total length of the line is checked not to exceed
        their respective maximum size.

        Args:
            keyword (str): 8-character keyword.
            text (str): value of the keyword.
            termination (str, optional): comment associated with the line.
        """
        keyword = keyword.ljust(LINE_KEYWORDS_SIZE)

        assert len(keyword) == LINE_KEYWORDS_SIZE

        line = '{0}({1}){2}'.format(keyword, text, termination)

        assert len(line) <= LINE_SIZE

        return line

    def _create_expline(self, keyword, value, termination=''):
        """
        Creates a exponent line.
        This type of line is characterised by a keyword, a value express as an
        exponent (see :func:`_toexponent`) and a termination string.
        The keyword and the total length of the line is checked not to exceed
        their respective maximum size.
        
        Args:
            keyword (str): 8-character keyword.
            value (float): value of the keyword.
            termination (str, optional): comment associated with the line.
        """
        keyword = keyword.rjust(LINE_KEYWORDS_SIZE)

        assert len(keyword) == LINE_KEYWORDS_SIZE

        line = '{0}({1},{2:4d}){3}'.format(keyword, _toexponent(value),
                                           0, termination)

        assert len(line) <= LINE_SIZE

        return line



"""
Input of PENEPMA simulation.
"""

# Standard library modules.

# Third party modules.

# Local modules.
from pypenelopetools.penelope.input import _PenelopeInputBase
import pypenelopetools.penelope.keywords as penelope_keywords
import pypenelopetools.penelope.separators as penelope_separators
import pypenelopetools.penepma.keywords as penepma_keywords
import pypenelopetools.penepma.separators as penepma_separators

# Globals and constants variables.

class PenepmaInput(_PenelopeInputBase):
    """
    Creates an object representing a PENEPMA input file.
    
    Attributes:
        TITLE (:class:`TITLE <pypenelopetools.penelope.keywords.TITLE>`): 
            Title of the job (up to 65 characters).
        SKPAR (:class:`SKPAR <pypenelopetools.penelope.keywords.SKPAR>`):
            Type of primary particle KPARP (1=electrons, 2=photons or 3=positrons).
        SENERG (:class:`SENERG <pypenelopetools.penelope.keywords.SENERG>`):
            Initial energy SE0 of primary particles.
        SPOSIT (:class:`SPOSIT <pypenelopetools.penelope.keywords.SPOSIT>`):
            Coordinates of the source centre.
        SRADI (:class:`SRADI <pypenelopetools.penepma.keywords.SRADI>`):
            Initial position of the particle is sampled randomly within a circle.
        SDIREC (:class:`SDIREC <pypenelopetools.penepma.keywords.SDIREC>`):
            Polar and azimuthal angles of the electron beam axis direction.
        SAPERT (:class:`SAPERT <pypenelopetools.penepma.keywords.SAPERT>`):
            Angular aperture of the electron beam.
        materials (:class:`Materials <pypenelopetools.penelope.keywords.Materials>`):
            Definition of materials.
        GEOMFN (:class:`GEOMFN <pypenelopetools.penelope.keywords.GEOMFN>`):
            Name of geometry definition file.
        DSMAX (:class:`DSMAX <pypenelopetools.penelope.keywords.DSMAX>`):
            Maximum step length of electrons and positrons in body.
        IFORCE (:class:`IFORCE <pypenelopetools.penelope.keywords.IFORCE>`):
            Forcing of interactions.
        IBRSPL (:class:`IBRSPL <pypenelopetools.penelope.keywords.IBRSPL>`):
            Bremsstrahlung splitting for electrons and positrons.
        IXRSPL (:class:`IXRSPL <pypenelopetools.penelope.keywords.IXRSPL>`):
            Splitting of characteristic x rays emitted.
        NBE (:class:`NBE <pypenelopetools.penelope.keywords.NBE>`):
            Definition of energy distributions of emerging particles.
        NBANGL (:class:`NBANGL <pypenelopetools.penelope.keywords.NBANGL>`):
            Definition of angular distributions of emerging particles.
        photon_detectors (:class:`PhotonDetectors <pypenelopetools.penepma.keywords.PhotonDetectors>`):
            Definition of the photon detectors.
        GRIDX (:class:`GRIDX <pypenelopetools.penelope.keywords.GRIDX>`):
            Definition of x-coordinates of the vertices of the dose box.
        GRIDY (:class:`GRIDY <pypenelopetools.penelope.keywords.GRIDY>`):
            Definition of y-coordinates of the vertices of the dose box.
        GRIDZ (:class:`GRIDZ <pypenelopetools.penelope.keywords.GRIDZ>`):
            Definition of z-coordinates of the vertices of the dose box.
        XRAYE (:class:`XRAYE <pypenelopetools.penepma.keywords.XRAYE>`):
            Space distribution of emission sites of x rays within an energy interval.
        XRLINE (:class:`XRLINE <pypenelopetools.penepma.keywords.XRLINE>`):
            Space distribution of emission sites of x rays.
        RESUME (:class:`RESUME <pypenelopetools.penelope.keywords.RESUME>`):
            Name of the resume file.
        DUMPTO (:class:`DUMPTO <pypenelopetools.penelope.keywords.DUMPTO>`):
            Name of dump file.
        DUMPP (:class:`DUMPP <pypenelopetools.penelope.keywords.DUMPP>`):
            Dump interval.
        RSEED (:class:`RSEED <pypenelopetools.penelope.keywords.RSEED>`):
            Seeds of the random-number generator.
        REFLIN (:class:`REFLIN <pypenelopetools.penepma.keywords.REFLIN>`):
            Termination of simulation based on relative statistical uncertainty of the intensity of line.
        NSIMSH (:class:`NSIMSH <pypenelopetools.penelope.keywords.NSIMSH>`):
            Desired number of simulated showers.
        TIME (:class:`TIME <pypenelopetools.penelope.keywords.TIME>`):
            Allotted simulation time.
    """

    def __init__(self):
        self.TITLE = penelope_keywords.TITLE()

        # Source definition
        self.SKPAR = penelope_keywords.SKPAR()
        self.SENERG = penelope_keywords.SENERG()
        self.SPOSIT = penelope_keywords.SPOSIT()
        self.SRADI = penepma_keywords.SRADI()
        self.SDIREC = penepma_keywords.SDIREC()
        self.SAPERT = penepma_keywords.SAPERT()

        # Material data and simulation parameters
        self.materials = penelope_keywords.Materials()

        # Geometry and local simulation parameters
        self.GEOMFN = penelope_keywords.GEOMFN()
        self.DSMAX = penelope_keywords.DSMAX()

        # Interaction forcing
        self.IFORCE = penelope_keywords.IFORCE()

        # Bremsstrahlung splitting
        self.IBRSPL = penelope_keywords.IBRSPL()

        # X-ray splitting
        self.IXRSPL = penelope_keywords.IXRSPL()

        # Emerging particles
        self.NBE = penelope_keywords.NBE()
        self.NBANGL = penelope_keywords.NBANGL()

        # Photon detectors
        self.photon_detectors = penepma_keywords.PhotonDetectors()

        # Spatial distribution
        self.GRIDX = penelope_keywords.GRIDX()
        self.GRIDY = penelope_keywords.GRIDY()
        self.GRIDZ = penelope_keywords.GRIDZ()
        self.XRAYE = penepma_keywords.XRAYE()
        self.XRLINE = penepma_keywords.XRLINE()

        # Job properties
        self.RESUME = penelope_keywords.RESUME()
        self.DUMPTO = penelope_keywords.DUMPTO()
        self.DUMPP = penelope_keywords.DUMPP()
        self.RSEED = penelope_keywords.RSEED()
        self.REFLIN = penepma_keywords.REFLIN()
        self.NSIMSH = penelope_keywords.NSIMSH()
        self.TIME = penelope_keywords.TIME()

    def get_keywords(self):
        return [self.TITLE,

                penelope_separators.DOT,
                penelope_separators.SOURCE_DEFINITION,
                self.SKPAR, self.SENERG, self.SPOSIT, self.SRADI, self.SDIREC, self.SAPERT,

                penelope_separators.DOT,
                penelope_separators.MATERIAL,
                self.materials,

                penelope_separators.DOT,
                penepma_separators.GEOMETRY,
                self.GEOMFN, self.DSMAX,

                penelope_separators.DOT,
                penelope_separators.INTERACTION_FORCING,
                self.IFORCE,

                penelope_separators.DOT,
                penelope_separators.BREMSSTRAHLUNG_SPLITTING,
                self.IBRSPL,

                penelope_separators.DOT,
                penelope_separators.XRAY_SPLITTING,
                self.IXRSPL,

                penelope_separators.DOT,
                penelope_separators.EMERGING_PARTICLES,
                self.NBE, self.NBANGL,

                penelope_separators.DOT,
                penepma_separators.PHOTON_DETECTORS,
                self.photon_detectors,

                penelope_separators.DOT,
                penepma_separators.SPATIAL_DISTRIBUTION,
                self.GRIDX, self.GRIDY, self.GRIDZ, self.XRAYE, self.XRLINE,

                penelope_separators.DOT,
                penelope_separators.JOB_PROPERTIES,
                self.RESUME, self.DUMPTO, self.DUMPP,

                penelope_separators.DOT,
                self.RSEED, self.REFLIN, self.NSIMSH, self.TIME]



""""""

# Standard library modules.

# Third party modules.

# Local modules.
from pypenelopetools.penelope.separator import Separator

# Globals and constants variables.

GEOMETRY_LIST_START = Separator('>>>>>>>> Geometry definition list.', 'GSTART')
GEOMETRY_LIST_END = Separator('<<<<<<<< End of the geometry definition list.', 'GEND')
DSMAX_EABSB = Separator('>>>>>>>> Local maximum step lengths and absorption energies.')
WOODCOCK = Separator(">>>>>>>> Woodcock's delta-scattering method for photons.")
COUNTER_ARRAY = Separator('>>>>>>>> Counter array dimensions and PDF ranges.')
PARTICLE_POSITIONS = Separator('>>>>>>>> Particle positions at the lower and upper planes.')
DOSE_CHARGE_DISTRIBUTION = Separator('>>>>>>>> Dose and charge distributions.')



"""
Definition of geometrical transformations.
"""

# Standard library modules.
import os
import math

# Third party modules.

# Local modules.
from pypenelopetools.pengeom.base import _GeometryBase, LINE_SEPARATOR

# Globals and constants variables.

class Rotation(_GeometryBase):
    """
    Represents a rotation using 3 Euler angles (YZY).

    Args:
        omega_deg (float): Rotation around the z-axis (deg).
        theta_deg (float): Rotation around the y-axis (deg).
        phi_deg (float): Rotation around the new z-axis (deg).
    """

    def __init__(self, omega_deg=0.0, theta_deg=0.0, phi_deg=0.0):
        self.omega_deg = omega_deg
        self.theta_deg = theta_deg
        self.phi_deg = phi_deg

    def __repr__(self):
        return "<Rotation(omega={0:g} deg, theta={1:g} deg, phi={2:g} deg)>" \
                .format(self.omega_deg, self.theta_deg, self.phi_deg)

    def __str__(self):
        return '(omega={0:g} deg, theta={1:g} deg, phi={2:g} deg)' \
                .format(self.omega_deg, self.theta_deg, self.phi_deg)

    def _read(self, fileobj, material_lookup, surface_lookup, module_lookup):
        line = self._peek_next_line(fileobj)
        while line != LINE_SEPARATOR:
            keyword, value, termination = self._parse_expline(line)

            if termination.startswith('RAD'):
                value = math.degrees(value)

            if keyword == 'OMEGA=':
                self.omega_deg = value
            elif keyword == 'THETA=':
                self.theta_deg = value
            elif keyword == 'PHI=':
                self.phi_deg = value

            line = self._read_next_line(fileobj)

    def _write(self, fileobj, index_lookup):
        line = self._create_expline('OMEGA=', self.omega_deg, ' DEG          (DEFAULT=0.0)')
        fileobj.write(line + os.linesep)

        line = self._create_expline('THETA=', self.theta_deg, ' DEG          (DEFAULT=0.0)')
        fileobj.write(line + os.linesep)

        line = self._create_expline('PHI=', self.phi_deg, ' DEG          (DEFAULT=0.0)')
        fileobj.write(line + os.linesep)

    @property
    def omega_deg(self):
        """float: Rotation around the z-axis (deg).
        The value must be between 0 and 360.
        """
        return self._omega

    @omega_deg.setter
    def omega_deg(self, angle):
        if angle < 0 or angle > 360.0:
            raise ValueError("Angle ({0}) must be between [0,360].".format(angle))
        self._omega = angle

    @property
    def theta_deg(self):
        """float: Rotation around the y-axis (deg).
        The value must be between 0 and 360.
        """
        return self._theta

    @theta_deg.setter
    def theta_deg(self, angle):
        if angle < 0 or angle > 360:
            raise ValueError("Angle ({0}) must be between [0,360].".format(angle))
        self._theta = angle

    @property
    def phi_deg(self):
        """float: Rotation around the new z-axis (deg).
        The new z-axis refer to the axis after the omega and theta rotation were
        applied on the original coordinate system.
        The value must be between 0 and 360.
        """
        return self._phi

    @phi_deg.setter
    def phi_deg(self, angle):
        if angle < 0 or angle > 360:
            raise ValueError("Angle ({0}) must be between [0,360].".format(angle))
        self._phi = angle

class Shift(_GeometryBase):
    """
    Represents a translation in space.
    
    Args:
        x_cm (float): Translation along the x direction (cm).
        y_cm (float): Translation along the y direction (cm).
        z_cm (float): Translation along the z direction (cm).
    """

    def __init__(self, x_cm=0.0, y_cm=0.0, z_cm=0.0):
        self.x_cm = x_cm
        self.y_cm = y_cm
        self.z_cm = z_cm

    def __repr__(self):
        return "<Shift(x={0:g} cm, y={1:g} cm, z={2:g} cm)>" \
            .format(self.x_cm, self.y_cm, self.z_cm)

    def __str__(self):
        return "(x={0:g} cm, y={1:g} cm, z={2:g} cm)" \
            .format(self.x_cm, self.y_cm, self.z_cm)

    def _read(self, fileobj, material_lookup, surface_lookup, module_lookup):
        line = self._peek_next_line(fileobj)
        while line != LINE_SEPARATOR:
            keyword, value, _ = self._parse_expline(line)

            if keyword == 'X-SHIFT=':
                self.x_cm = value
            elif keyword == 'Y-SHIFT=':
                self.y_cm = value
            elif keyword == 'Z-SHIFT=':
                self.z_cm = value

            line = self._read_next_line(fileobj)

    def _write(self, fileobj, index_lookup):
        line = self._create_expline('X-SHIFT=', self.x_cm, '              (DEFAULT=0.0)')
        fileobj.write(line + os.linesep)

        line = self._create_expline('Y-SHIFT=', self.y_cm, '              (DEFAULT=0.0)')
        fileobj.write(line + os.linesep)

        line = self._create_expline('Z-SHIFT=', self.z_cm, '              (DEFAULT=0.0)')
        fileobj.write(line + os.linesep)

    @property
    def x_cm(self):
        """float: Translation along the x direction (cm)."""
        return self._x

    @x_cm.setter
    def x_cm(self, shift):
        self._x = shift

    @property
    def y_cm(self):
        """float: Translation along the y direction (cm)."""
        return self._y

    @y_cm.setter
    def y_cm(self, shift):
        self._y = shift

    @property
    def z_cm(self):
        """float: Translation along the z direction (cm)."""
        return self._z

    @z_cm.setter
    def z_cm(self, shift):
        self._z = shift

class Scale(_GeometryBase):
    """
    Represents scaling.
    
    Args:
        x (float): Scaling along the x direction.
        y (float): Scaling along the y direction.
        z (float): Scaling along the z direction.
    """

    def __init__(self, x=1.0, y=1.0, z=1.0):
        self.x = x
        self.y = y
        self.z = z

    def __repr__(self):
        return "<Shift(x={0:g}, y={1:g}, z={2:g})>" \
            .format(self.x, self.y, self.z)

    def __str__(self):
        return "(x={0:g}, y={1:g}, z={2:g})".format(self.x, self.y, self.z)

    def _read(self, fileobj, material_lookup, surface_lookup, module_lookup):
        line = self._peek_next_line(fileobj)
        while line != LINE_SEPARATOR:
            keyword, value, _ = self._parse_expline(line)

            if keyword == 'X-SCALE=':
                self.x = value
            elif keyword == 'Y-SCALE=':
                self.y = value
            elif keyword == 'Z-SCALE=':
                self.z = value

            line = self._read_next_line(fileobj)

    def _write(self, fileobj, index_lookup):
        line = self._create_expline('X-SCALE=', self.x, '              (DEFAULT=1.0)')
        fileobj.write(line + os.linesep)

        line = self._create_expline('Y-SCALE=', self.y, '              (DEFAULT=1.0)')
        fileobj.write(line + os.linesep)

        line = self._create_expline('Z-SCALE=', self.z, '              (DEFAULT=1.0)')
        fileobj.write(line + os.linesep)

    @property
    def x(self):
        """float: Scaling along the x direction.
        The value cannot be 0.
        """
        return self._x

    @x.setter
    def x(self, scale):
        if scale == 0.0:
            raise ValueError("X scale cannot be equal to 0.")
        self._x = scale

    @property
    def y(self):
        """float: Scaling along the y direction.
        The value cannot be 0.
        """
        return self._y

    @y.setter
    def y(self, scale):
        if scale == 0.0:
            raise ValueError("Y scale cannot be equal to 0.")
        self._y = scale

    @property
    def z(self):
        """float: Scaling along the z direction.
        The value cannot be 0.
        """
        return self._z

    @z.setter
    def z(self, scale):
        if scale == 0.0:
            raise ValueError("Z scale cannot be equal to 0.")
        self._z = scale



"""
Definition of module.
"""

# Standard library modules.
import os
import enum

# Third party modules.

# Local modules.
from pypenelopetools.pengeom.transformation import Rotation, Shift
from pypenelopetools.pengeom.mixin import DescriptionMixin, ModuleMixin
from pypenelopetools.pengeom.base import _GeometryBase, LINE_EXTRA, LINE_SEPARATOR
from pypenelopetools.material import VACUUM

# Globals and constants variables.

class SidePointer(enum.IntEnum):
    """
    Whether the surface is pointing in the positive or negative direction.
    """

    POSITIVE = 1
    """Positive direction."""

    NEGATIVE = -1
    """Negative direction."""

class Module(DescriptionMixin, ModuleMixin, _GeometryBase):
    """
    Definition of a module.
    
    Args:
        material (:obj:`Material <pypenelopetools.material.Material>`, optional):
            Material associated with this module.
            If ``None``, the material is set to 
            :obj:`VACUUM <pypenelopetools.material.VACUUM>`.
        description (str): Description of the module
    """

    def __init__(self, material=None, description=''):
        if material is None:
            material = VACUUM
        self.material = material

        self.description = description

        self._surfaces = {}
        self._modules = set()

        self._rotation = Rotation()
        self._shift = Shift()

    def __repr__(self):
        return '<Module(description={0}, material={1}, surfaces_count={2:d}, modules_count={3:d}, rotation={4}, shift={5})>' \
            .format(self.description, self.material, len(self._surfaces),
                    len(self._modules), str(self.rotation), str(self.shift))

    def _read(self, fileobj, material_lookup, surface_lookup, module_lookup):
        line = self._read_next_line(fileobj)
        _, _, self.description = self._parse_line(line)

        line = self._read_next_line(fileobj)
        keyword, material_index, _ = self._parse_line(line)
        if keyword != 'MATERIAL':
            raise IOError('Expected keyword "MATERIAL" instead of "{0}"'.format(keyword))

        material_index = int(material_index)
        if material_index not in material_lookup:
            raise IOError('No material {0} in lookup table'.format(material_index))
        self.material = material_lookup[material_index]

        line = self._read_next_line(fileobj)
        while line != LINE_EXTRA and line != LINE_SEPARATOR:
            keyword, index, termination = self._parse_line(line)
            index = int(index)

            if keyword == 'SURFACE':
                pointer = int(termination[16:18])
                surface = surface_lookup[index]
                self.add_surface(surface, pointer)

            elif keyword == 'MODULE':
                submodule = module_lookup[index]
                self.add_module(submodule)

            else:
                raise IOError('Unknown keyword: {0}'.format(keyword))

            line = self._read_next_line(fileobj)

        if line == LINE_EXTRA:
            extra_offset = fileobj.tell()
            self.rotation._read(fileobj, material_lookup, surface_lookup, module_lookup)

            fileobj.seek(extra_offset)
            self.shift._read(fileobj, material_lookup, surface_lookup, module_lookup)

    def _write(self, fileobj, index_lookup):
        index = index_lookup[self]
        text = "{:4d}".format(index)
        termination = " " + self.description
        line = self._create_line('MODULE', text, termination)
        fileobj.write(line + os.linesep)

        # Material index
        index = index_lookup[self.material]
        text = "{:4d}".format(index)
        line = self._create_line('MATERIAL', text)
        fileobj.write(line + os.linesep)

        # Surface pointers
        surfaces = sorted((index_lookup[surface], surface, pointer)
                          for surface, pointer in self._surfaces.items())

        for index, surface, pointer in surfaces:
            text = "{:4d}".format(index)
            termination = ", SIDE POINTER=({:2d})".format(pointer)
            line = self._create_line('SURFACE', text, termination)
            fileobj.write(line + os.linesep)

        # Module indexes
        modules = sorted((index_lookup[module], module)
                          for module in self.get_modules())

        for index, module in modules:
            text = "{:4d}".format(index)
            line = self._create_line('MODULE', text)
            fileobj.write(line + os.linesep)

        # Separator
        fileobj.write(LINE_EXTRA + os.linesep)

        # Rotation
        self.rotation._write(fileobj, index_lookup)

        # Shift
        self.shift._write(fileobj, index_lookup)

        fileobj.write(LINE_SEPARATOR + os.linesep)

    def add_surface(self, surface, pointer):
        """
        Adds a surface.
        
        Args:
            surface (:obj:`SurfaceImplicit <pypenelopetools.pengeom.surface.SurfaceImplicit>` or :obj:`SurfaceReduced <pypenelopetools.pengeom.surface.SurfaceReduced>`):
                Surface to add.
            pointer (:obj:`SidePointer`): 
                Whether the surface is pointing in the positive or negative 
                direction.
        """
        if isinstance(pointer, int):
            if pointer == SidePointer.NEGATIVE:
                pointer = SidePointer.NEGATIVE
            elif pointer == SidePointer.POSITIVE:
                pointer = SidePointer.POSITIVE
        if pointer not in SidePointer:
            raise ValueError("Pointer ({0}) must be either -1 or 1.".format(pointer))
        if surface in self._surfaces:
            raise ValueError("Module already contains this surface.")
        self._surfaces[surface] = pointer

    def pop_surface(self, surface):
        """
        Removes a surface.
        
        Args:
            surface (:obj:`SurfaceImplicit <pypenelopetools.pengeom.surface.SurfaceImplicit>` or :obj:`SurfaceReduced <pypenelopetools.pengeom.surface.SurfaceReduced>`):
                Surface to remove.
        """
        self._surfaces.pop(surface)

    def clear_surfaces(self):
        """
        Clear all surfaces.
        """
        self._surfaces.clear()

    def get_surface_pointer(self, surface):
        """
        Returns the surface pointer for the specified surface.
        
        Args:
            surface (:obj:`SurfaceImplicit <pypenelopetools.pengeom.surface.SurfaceImplicit>` or :obj:`SurfaceReduced <pypenelopetools.pengeom.surface.SurfaceReduced>`):
                Surface of interest.
        
        Returns:
            :obj:`SidePointer`: Side pointer.
        """
        return self._surfaces[surface]

    def get_surfaces(self):
        """
        Returns:
            tuple: All surfaces.
        """
        return tuple(self._surfaces.keys())

    @property
    def rotation(self):
        """:obj:`Rotation <pypenelopetools.pengeom.transformation.Rotation>`: Rotation of the module."""
        return self._rotation

    @property
    def shift(self):
        """:obj:`Shift <pypenelopetools.pengeom.transformation.Shift>`: Shift/translation of the module."""
        return self._shift



"""
Results of PENEPMA simulation.
"""

# Standard library modules.
import os

# Third party modules.
from uncertainties import ufloat, unumpy

import pyxray

# Local modules.
from pypenelopetools.penelope.result import _PenelopeResultBase

# Globals and constants variables.

class PenepmaResult(_PenelopeResultBase):
    """
    Results from ``penepma-res.dat``.
    
    .. note::
       All results are expressed using the 
       `uncertainties <https://pythonhosted.org/uncertainties>`_ package.
       The nominal value can be accessed with the property ``nominal_value`` or
       the abbreviation ``n``, whereas the standard deviation (1-sigma), with 
       the property ``std_dev`` or ``s``.
       For example::
       
           result.simulation_time_s.n #-> 100.0
           result.simulation_time_s.s #-> 0.0
    
    Attributes:
        simulation_time_s (ufloat):
            Simulation time in seconds.
        simulation_speed_1_per_s (ufloat):
            Simulation speed in simulation per second.
        simulated_primary_showers (ufloat): 
            Number of primary showers simulated.
            
        upbound_primary_particles (ufloat):
            Number of primary particles that exited the geometry upwards.
        downbound_primary_particles (ufloat):
            Number of primary particles that exited the geometry downwards.
        absorbed_primary_particles (ufloat):
            Number of primary particles that were absorbed within the geometry.
            
        upbound_fraction (ufloat):
            Fraction of primary particles that exited the geometry upwards.
        downbound_fraction (ufloat):
            Fraction of primary particles that exited the geometry downwards.
        absorbed_fraction (ufloat):
            Fraction of primary particles that were absorbed within the geometry.
            
        upbound_secondary_electron_generation_probabilities (ufloat):
            Probability of second generation electrons exited the geometry upwards.
        downbound_secondary_electron_generation_probabilities (ufloat):
            Probability of second generation electrons exited the geometry downwards.
        absorbed_secondary_electron_generation_probabilities (ufloat):
            Probability of second generation electrons absorbed within the geometry.
        upbound_secondary_photon_generation_probabilities (ufloat):
            Probability of second generation photons exited the geometry upwards.
        downbound_secondary_photon_generation_probabilities (ufloat):
            Probability of second generation photons exited the geometry downwards.
        absorbed_secondary_photon_generation_probabilities (ufloat):
            Probability of second generation photons absorbed within the geometry.
        upbound_secondary_positron_generation_probabilities (ufloat):
            Probability of second generation positrons exited the geometry upwards.
        downbound_secondary_positron_generation_probabilities (ufloat):
            Probability of second generation positrons exited the geometry downwards.
        absorbed_secondary_positron_generation_probabilities (ufloat):
            Probability of second generation positrons absorbed within the geometry.

        average_deposited_energy_eV (dict(int, ufloat)):
            Average deposited energy in each body.
            Dictionary where keys are indexes of body and values, the average
            deposited energy in eV.
        average_photon_energy_eV (dict(int, ufloat)):
            Average photon energy in each detector.
            Dictionary where keys are indexes of detector and values, the 
            average energy in eV.

        last_random_seed1 (ufloat):
            Last first seed of the random number generator.
        last_random_seed2 (ufloat):
            Last second seed of the random number generator.
        reference_line_uncertainty (ufloat):
            Relative uncertainty of the x-ray line used as a termination condition
    """

    def __init__(self):
        super().__init__()

        self.simulation_time_s = ufloat(0.0, 0.0)
        self.simulation_speed_1_per_s = ufloat(0.0, 0.0)
        self.simulated_primary_showers = ufloat(0.0, 0.0)

        self.upbound_primary_particles = ufloat(0.0, 0.0)
        self.downbound_primary_particles = ufloat(0.0, 0.0)
        self.absorbed_primary_particles = ufloat(0.0, 0.0)

        self.upbound_fraction = ufloat(0.0, 0.0)
        self.downbound_fraction = ufloat(0.0, 0.0)
        self.absorbed_fraction = ufloat(0.0, 0.0)

        self.upbound_secondary_electron_generation_probabilities = ufloat(0.0, 0.0)
        self.downbound_secondary_electron_generation_probabilities = ufloat(0.0, 0.0)
        self.absorbed_secondary_electron_generation_probabilities = ufloat(0.0, 0.0)
        self.upbound_secondary_photon_generation_probabilities = ufloat(0.0, 0.0)
        self.downbound_secondary_photon_generation_probabilities = ufloat(0.0, 0.0)
        self.absorbed_secondary_photon_generation_probabilities = ufloat(0.0, 0.0)
        self.upbound_secondary_positron_generation_probabilities = ufloat(0.0, 0.0)
        self.downbound_secondary_positron_generation_probabilities = ufloat(0.0, 0.0)
        self.absorbed_secondary_positron_generation_probabilities = ufloat(0.0, 0.0)

        self.average_deposited_energy_eV = {}
        self.average_photon_energy_eV = {}

        self.last_random_seed1 = ufloat(0.0, 0.0)
        self.last_random_seed2 = ufloat(0.0, 0.0)

        self.reference_line_uncertainty = ufloat(0.0, 0.0)

    def read(self, fileobj):
        line = self._read_until_line_startswith(fileobj, 'Simulation time')
        val, = self._read_all_values(line)
        self.simulation_time_s = ufloat(val, 0.0)

        line = self._read_until_line_startswith(fileobj, 'Simulation speed')
        val, = self._read_all_values(line)
        self.simulation_speed_1_per_s = ufloat(val, 0.0)

        line = self._read_until_line_startswith(fileobj, 'Simulated primary showers')
        val, = self._read_all_values(line)
        self.simulated_primary_showers = ufloat(val, 0.0)

        line = self._read_until_line_startswith(fileobj, 'Upbound primary particles')
        val, = self._read_all_values(line)
        self.upbound_primary_particles = ufloat(val, 0.0)

        line = self._read_until_line_startswith(fileobj, 'Downbound primary particles')
        val, = self._read_all_values(line)
        self.downbound_primary_particles = ufloat(val, 0.0)

        line = self._read_until_line_startswith(fileobj, 'Absorbed primary particles')
        val, = self._read_all_values(line)
        self.absorbed_primary_particles = ufloat(val, 0.0)

        line = self._read_until_line_startswith(fileobj, 'Upbound fraction')
        val, unc = self._read_all_values(line)
        self.upbound_fraction = ufloat(val, unc / 3)

        line = self._read_until_line_startswith(fileobj, 'Downbound fraction')
        val, unc = self._read_all_values(line)
        self.downbound_fraction = ufloat(val, unc / 3)

        line = self._read_until_line_startswith(fileobj, 'Absorption fraction')
        val, unc = self._read_all_values(line)
        self.absorbed_fraction = ufloat(val, unc / 3)

        self._read_until_line_startswith(fileobj, 'Secondary-particle generation probabilities')
        fileobj.readline() # skip header
        fileobj.readline() # skip header

        fileobj.readline() # skip header
        val_el, val_ph, val_po = self._read_all_values(fileobj.readline())
        unc_el, unc_ph, unc_po = self._read_all_values(fileobj.readline())
        self.upbound_secondary_electron_generation_probabilities = ufloat(val_el, unc_el / 3)
        self.upbound_secondary_photon_generation_probabilities = ufloat(val_ph, unc_ph / 3)
        self.upbound_secondary_positron_generation_probabilities = ufloat(val_po, unc_po / 3)

        fileobj.readline() # skip header
        val_el, val_ph, val_po = self._read_all_values(fileobj.readline())
        unc_el, unc_ph, unc_po = self._read_all_values(fileobj.readline())
        self.downbound_secondary_electron_generation_probabilities = ufloat(val_el, unc_el / 3)
        self.downbound_secondary_photon_generation_probabilities = ufloat(val_ph, unc_ph / 3)
        self.downbound_secondary_positron_generation_probabilities = ufloat(val_po, unc_po / 3)

        fileobj.readline() # skip header
        val_el, val_ph, val_po = self._read_all_values(fileobj.readline())
        unc_el, unc_ph, unc_po = self._read_all_values(fileobj.readline())
        self.absorbed_secondary_electron_generation_probabilities = ufloat(val_el, unc_el / 3)
        self.absorbed_secondary_photon_generation_probabilities = ufloat(val_ph, unc_ph / 3)
        self.absorbed_secondary_positron_generation_probabilities = ufloat(val_po, unc_po / 3)

        self.average_deposited_energy_eV.clear()
        self._read_until_line_startswith(fileobj, 'Average deposited energies (bodies)')
        line = fileobj.readline().strip()
        while line.startswith('Body'):
            body = int(line[5:9])
            val, unc, _effic = self._read_all_values(line)
            self.average_deposited_energy_eV[body] = ufloat(val, unc / 3)
            line = fileobj.readline().strip()

        self.average_photon_energy_eV.clear()
        self._read_until_line_startswith(fileobj, 'Average photon energy at the detectors')
        line = fileobj.readline().strip()
        while line.startswith('Detector'):
            detector = int(line[10:12])
            val, unc, _effic = self._read_all_values(line)
            self.average_photon_energy_eV[detector] = ufloat(val, unc / 3)
            line = fileobj.readline().strip()

        line = self._read_until_line_startswith(fileobj, 'Last random seeds')
        seed1, seed2 = map(int, line.split('=')[1].split(','))
        self.last_random_seed1 = ufloat(seed1, 0.0)
        self.last_random_seed2 = ufloat(seed2, 0.0)

        try:
            self._read_until_line_startswith(fileobj, 'Reference line')
            val, = self._read_all_values(fileobj.readline())
            self.reference_line_uncertainty = ufloat(val, 0.0)
        except IOError:
            self.reference_line_uncertainty = ufloat(0.0, 0.0)

    def read_directory(self, dirpath):
        filepath = os.path.join(dirpath, 'penepma-res.dat')
        with open(filepath, 'r') as fp:
            self.read(fp)

class _PenepmaPhotonDetectorResult(_PenelopeResultBase):
    """
    Base result associated with a photon detector.
    
    Args:
        detector_index (int): Index of the detector to read the results from.
        
    Attributes:
        detector_index (int):
            Index of detector
        theta1_deg (ufloat): 
            Lower limit polar angle in deg.
        theta2_deg (ufloat): 
            Upper limit polar angle in deg.
        phi1_deg (ufloat): 
            Lower limit azimuthal angle in deg.
        phi2_deg (ufloat): 
            Upper limit azimuthal angle in deg.
    """

    def __init__(self, detector_index):
        super().__init__()
        self.detector_index = detector_index
        self.theta1_deg = ufloat(0.0, 0.0)
        self.theta2_deg = ufloat(0.0, 0.0)
        self.phi1_deg = ufloat(0.0, 0.0)
        self.phi2_deg = ufloat(0.0, 0.0)

    def read(self, fileobj):
        line = self._read_until_line_startswith(fileobj, '#  Results from PENEPMA. Output from photon detector')
        _, detector_index = line.rsplit('#', 1)
        detector_index = int(detector_index)
        if detector_index != self.detector_index:
            raise IOError('Mismatch of detector index (expected {}, found {})'
                          .format(self.detector_index, detector_index))

        line = self._read_until_line_startswith(fileobj, '#  Angular intervals :')
        theta1_deg, theta2_deg = self._read_all_values(line)
        self.theta1_deg = ufloat(theta1_deg, 0.0)
        self.theta2_deg = ufloat(theta2_deg, 0.0)

        line = fileobj.readline()
        phi1_deg, phi2_deg = self._read_all_values(line)
        self.phi1_deg = ufloat(phi1_deg, 0.0)
        self.phi2_deg = ufloat(phi2_deg, 0.0)

class PenepmaIntensityResult(_PenepmaPhotonDetectorResult):
    """
    Results from ``pe-intens-XX.dat``, where ``XX`` is the index of the detector.
    The intensities are given for each characteristic x-ray detected by
    the detector.
     
    .. note::
       The characteristic x-rays are expressed using :class:`XrayLine` object
       of the `pyxray <https://github.com/openmicroanalysis/pyxray>`_ package.
       :class:`XrayLine` objects define the atomic number and x-ray transition 
       of characteristic x-rays.
       
       The intensities are expressed using the 
       `uncertainties <https://pythonhosted.org/uncertainties>`_ package.
       The nominal value can be accessed with the property ``nominal_value`` or
       the abbreviation ``n``, whereas the standard deviation (1-sigma), with 
       the property ``std_dev`` or ``s``.
       
       For example::
       
           import pyxray
           x = pyxray.XrayLine(29, 'Ka1')
           result.total_intensities_1_per_sr_electron[x].n #-> 8.56e-10
           result.total_intensities_1_per_sr_electron[x].s #-> 0.15e-10
    
    Args:
        detector_index (int): Index of the detector to read the results from.
        
    Attributes:
        detector_index (int):
            Index of detector
        theta1_deg (ufloat): 
            Lower limit polar angle in deg.
        theta2_deg (ufloat): 
            Upper limit polar angle in deg.
        phi1_deg (ufloat): 
            Lower limit azimuthal angle in deg.
        phi2_deg (ufloat): 
            Upper limit azimuthal angle in deg.
            
        primary_intensities_1_per_sr_electron (dict(XrayLine, ufloat)):
            Intensities of characteristic x-rays generated by primary electrons
            and measured by the detector.
        characteristic_fluorescence_intensities_1_per_sr_electron (dict(XrayLine, ufloat)):
            Intensities of characteristic x-rays generated by the fluorescence
            of characteristic x-rays and measured by the detector.
        bremsstrahlung_fluorescence_intensities_1_per_sr_electron (dict(XrayLine, ufloat)):
            Intensities of characteristic x-rays generated by the fluorescence
            of Bremsstrahlung x-rays and measured by the detector.
        total_fluorescence_intensities_1_per_sr_electron (dict(XrayLine, ufloat)):
            Intensities of characteristic x-rays generated by fluorescence
            (characteristic and Bremsstrahlung) and measured by the detector.
            Intensities are equal to the sum of 
            *characteristic_fluorescence_intensities_1_per_sr_electron* and
            *bremsstrahlung_fluorescence_intensities_1_per_sr_electron*.
        total_intensities_1_per_sr_electron (dict(XrayLine, ufloat)):
            Intensities of characteristic x-rays generated and measured by the 
            detector.
            Intensities are equal to the sum of 
            *primary_intensities_1_per_sr_electron* and
            *total_fluorescence_intensities_1_per_sr_electron*.
    """

    def __init__(self, detector_index):
        super().__init__(detector_index)

        self.primary_intensities_1_per_sr_electron = {}
        self.characteristic_fluorescence_intensities_1_per_sr_electron = {}
        self.bremsstrahlung_fluorescence_intensities_1_per_sr_electron = {}
        self.total_fluorescence_intensities_1_per_sr_electron = {}
        self.total_intensities_1_per_sr_electron = {}

    def read(self, fileobj):
        super().read(fileobj)

        self.primary_intensities_1_per_sr_electron.clear()
        self.characteristic_fluorescence_intensities_1_per_sr_electron.clear()
        self.bremsstrahlung_fluorescence_intensities_1_per_sr_electron.clear()
        self.total_fluorescence_intensities_1_per_sr_electron.clear()
        self.total_intensities_1_per_sr_electron.clear()
        self._read_until_line_startswith(fileobj, '# IZ S0 S1  E (eV)')
        for line in fileobj:
            z = int(line[3:5])
            dst = pyxray.atomic_subshell(line[6:8].strip())
            src = pyxray.atomic_subshell(line[9:11].strip())
            xrayline = pyxray.xray_line(z, (src, dst))

            _e, val_p, unc_p, val_c, unc_c, val_b, unc_b, val_tf, unc_tf, val_t, unc_t = self._read_all_values(line)
            self.primary_intensities_1_per_sr_electron[xrayline] = ufloat(val_p, unc_p / 3)
            self.characteristic_fluorescence_intensities_1_per_sr_electron[xrayline] = ufloat(val_c, unc_c / 3)
            self.bremsstrahlung_fluorescence_intensities_1_per_sr_electron[xrayline] = ufloat(val_b, unc_b / 3)
            self.total_fluorescence_intensities_1_per_sr_electron[xrayline] = ufloat(val_tf, unc_tf / 3)
            self.total_intensities_1_per_sr_electron[xrayline] = ufloat(val_t, unc_t / 3)

    def read_directory(self, dirpath):
        filepath = os.path.join(dirpath, 'pe-intens-{:02d}.dat'.format(self.detector_index))
        with open(filepath, 'r') as fp:
            self.read(fp)

class PenepmaSpectrumResult(_PenepmaPhotonDetectorResult):
    """
    Results from ``pe-spect-XX.dat``, where ``XX`` is the index of the detector.
    The spectrum is stored as a `numpy <http://numpy.org>`_ array where the
    first column contains the energies in eV and the second the intensities
    in 1/(sr.electron).
    
    .. note::
       The energies and intensities are expressed using the 
       `uncertainties <https://pythonhosted.org/uncertainties>`_ package.
       The nominal value can be accessed with the property ``nominal_value`` or
       the abbreviation ``n``, whereas the standard deviation (1-sigma), with 
       the property ``std_dev`` or ``s``.
       
       For example::
           
           from uncertainty import unumpy
           energies_eV = unumpy.nominal_values(result.spectrum[:,0])
           intensities = unumpy.nominal_values(result.spectrum[:,1])
    
    Args:
        detector_index (int): Index of the detector to read the results from.
        
    Attributes:
        detector_index (int):
            Index of detector
        theta1_deg (ufloat): 
            Lower limit polar angle in deg.
        theta2_deg (ufloat): 
            Upper limit polar angle in deg.
        phi1_deg (ufloat): 
            Lower limit azimuthal angle in deg.
        phi2_deg (ufloat): 
            Upper limit azimuthal angle in deg.
            
        energy_window_start_eV (ufloat):
            Energy of first window in eV.
        energy_window_end_eV (ufloat):
            Energy of last window in eV.
        channel_width_eV (ufloat):
            Width of one channel in eV.

        spectrum (unumpy.uarray):
            Array where the first column contains the energies in eV and 
            the second the intensities measured by the detector in 
            1/(sr.electron).
    """

    def __init__(self, detector_index):
        super().__init__(detector_index)

        self.energy_window_start_eV = ufloat(0.0, 0.0)
        self.energy_window_end_eV = ufloat(0.0, 0.0)
        self.channel_width_eV = ufloat(0.0, 0.0)

        self.spectrum = unumpy.uarray([], [])

    def read(self, fileobj):
        super().read(fileobj)

        line = self._read_until_line_startswith(fileobj, '#  Energy window =')
        start_eV, end_eV = self._read_all_values(line)
        self.energy_window_start_eV = ufloat(start_eV, 0.0)
        self.energy_window_end_eV = ufloat(end_eV, 0.0)

        line = self._read_until_line_startswith(fileobj, '#  Channel width =')
        channel_width_eV, = self._read_all_values(line)
        self.channel_width_eV = ufloat(channel_width_eV, 0.0)

        spectrum = []
        spectrum_unc = []
        self._read_until_end_of_comments(fileobj)
        for line in fileobj:
            energy_eV, val, unc = self._read_all_values(line)
            spectrum.append([energy_eV, val])
            spectrum_unc.append([0.0, unc / 3])

        self.spectrum = unumpy.uarray(spectrum, spectrum_unc)

    def read_directory(self, dirpath):
        filepath = os.path.join(dirpath, 'pe-spect-{:02d}.dat'.format(self.detector_index))
        with open(filepath, 'r') as fp:
            self.read(fp)

    @property
    def energies_eV(self):
        """numpy array: Nominal values of the energy axis in eV."""
        return unumpy.nominal_values(self.spectrum[:, 0])

    @property
    def intensities_1_per_sr_electron(self):
        """numpy array: Nominal values of the intensity axis in 1/(sr.electron)."""
        return unumpy.nominal_values(self.spectrum[:, 1])



#!/usr/bin/env python
""" """

# Standard library modules.
import unittest
import logging
import os
import io

# Third party modules.

# Local modules.
from pypenelopetools.pencyl.input import PencylInput
from pypenelopetools.material import Material

# Globals and constants variables.

def create_example1_disc():
    # Create materials
    material_cu = Material('Cu', {29: 1.0}, 8.9)

    # Create input
    input = PencylInput()

    input.TITLE.set('Point source and a homogeneous cylinder.')

    definition = input.geometry_definitions.add(0.0, 0.005)
    definition.CYLIND.add(1, 0, 0.01)

    input.SKPAR.set(1)
    input.SENERG.set(40e3)
    input.SPOSIT.set(0, 0, -0.0001)
    input.SCONE.set(0, 0, 5)

    input.materials.add(1, material_cu.filename, 1e3, 1e3, 1e3, 0.05, 0.05, 1e3, 1e3)

    input.DSMAX.add(1, 1, 1e-4)

    input.IFORCE.add(1, 1, 1, 4, 2000, 0.1, 2.0)
    input.IFORCE.add(1, 1, 1, 5, 200, 0.1, 2.0)

    input.IBRSPL.add(1, 1, 2)

    input.IXRSPL.add(1, 1, 2)

    input.NBE.set(0, 0, 100)
    input.NBANGL.set(45, 18)

    input.EMERGP.set(0.005, 100)

    detector = input.energy_deposition_detectors.add(0, 0, 100)
    detector.EDBODY.add(1, 1)

    input.DOSE2D.add(1, 1, 100, 50)

    input.RESUME.set('dump.dat')
    input.DUMPTO.set('dump.dat')
    input.DUMPP.set(60)

    input.RSEED.set(1, 1)
    input.NSIMSH.set(1e9)
    input.TIME.set(600)

    return input

def create_example3_detector():
    # Create materials
    material_nai = Material('NaI', {11: 0.1534, 53: 0.8466}, 3.667)
    material_al2o3 = Material('Al2O3', {8: 0.4707, 13: 0.5293}, 3.97)
    material_al = Material('Al', {13: 1.0}, 2.7)
    materials = (material_nai, material_al2o3, material_al)

    # Create input
    input = PencylInput()

    input.TITLE.set('NaI detector with Al cover and Al2O3 reflecting foil')

    layer1 = input.geometry_definitions.add(-0.24, -0.16, 0.0, 0.0)
    layer1.CYLIND.add(3, 0.00, 4.05)

    layer2 = input.geometry_definitions.add(-0.16, 0.00)
    layer2.CYLIND.add(2, 0.00, 3.97)
    layer2.CYLIND.add(3, 3.97, 4.05)

    layer3 = input.geometry_definitions.add(0.00, 7.72)
    layer3.CYLIND.add(1, 0.00, 3.81)
    layer3.CYLIND.add(2, 3.81, 3.97)
    layer3.CYLIND.add(3, 3.97, 4.05)

    layer4 = input.geometry_definitions.add(7.72, 9.72)
    layer4.CYLIND.add(3, 0.00, 4.05)

    input.SKPAR.set(2)
    input.SENERG.set(9.5e6)
    input.SPOSIT.set(0, 0, -10.0)
    input.SCONE.set(0, 0, 0)

    input.materials.add(1, material_nai.filename, 5.0e4, 5.0e3, 5.0e4, 0.1, 0.1, 2e3, 2e3)
    input.materials.add(2, material_al2o3.filename, 5.0e4, 5.0e3, 5.0e4, 0.1, 0.1, 2e3, 2e3)
    input.materials.add(3, material_al.filename, 5.0e4, 5.0e3, 5.0e4, 0.1, 0.1, 2e3, 2e3)

    detector = input.energy_deposition_detectors.add(0, 1e7, 1000)
    detector.EDBODY.add(3, 1)

    input.DOSE2D.add(3, 1, 50, 50)

    input.RESUME.set('dump.dat')
    input.DUMPTO.set('dump.dat')
    input.DUMPP.set(60)

    input.NSIMSH.set(1e8)
    input.TIME.set(2e9)

    return input, materials

class TestPencylInput(unittest.TestCase):

    def setUp(self):
        super().setUp()

        self.testdatadir = os.path.join(os.path.dirname(__file__),
                                        '..', 'testdata', 'pencyl')

    def _write_read_input(self, input):
        fileobj = io.StringIO()

        try:
            input.write(fileobj)

            fileobj.seek(0)
            outinput = PencylInput()
            outinput.read(fileobj)
        finally:
            fileobj.close()

        return outinput

    def _test_example1_disc(self, input):
        geometry_definitions, = input.geometry_definitions.get()
        self.assertEqual(1, len(geometry_definitions))

        zlow, zhigh, xcen, ycen, cylinders = geometry_definitions[0]
        self.assertAlmostEqual(0.0, zlow, 5)
        self.assertAlmostEqual(0.005, zhigh, 5)
        self.assertIsNone(xcen)
        self.assertIsNone(ycen)
        self.assertEqual(1, len(cylinders))

        material, rin, rout = cylinders[0]
        self.assertEqual(1, material)
        self.assertAlmostEqual(0.0, rin, 5)
        self.assertAlmostEqual(0.01, rout, 5)

        kparp, = input.SKPAR.get()
        self.assertEqual(1, kparp)

        se0, = input.SENERG.get()
        self.assertAlmostEqual(40e3, se0, 5)

        sx0, sy0, sz0 = input.SPOSIT.get()
        self.assertAlmostEqual(0.0, sx0, 5)
        self.assertAlmostEqual(0.0, sy0, 5)
        self.assertAlmostEqual(-0.0001, sz0, 5)

        theta, phi, alpha = input.SCONE.get()
        self.assertAlmostEqual(0.0, theta, 5)
        self.assertAlmostEqual(0.0, phi, 5)
        self.assertAlmostEqual(5.0, alpha, 5)

        materials, = input.materials.get()
        self.assertEqual(1, len(materials))

        filename, eabs1, eabs2, eabs3, c1, c2, wcc, wcr = materials[0]
        self.assertEqual('Cu.mat', filename)
        self.assertAlmostEqual(1e3, eabs1, 5)
        self.assertAlmostEqual(1e3, eabs2, 5)
        self.assertAlmostEqual(1e3, eabs3, 5)
        self.assertAlmostEqual(0.05, c1, 5)
        self.assertAlmostEqual(0.05, c2, 5)
        self.assertAlmostEqual(1e3, wcc, 5)
        self.assertAlmostEqual(1e3, wcr, 5)

        dsmaxs, = input.DSMAX.get()
        self.assertEqual(1, len(dsmaxs))

        kl, kc, dsmax = dsmaxs[0]
        self.assertEqual(1, kl)
        self.assertEqual(1, kc)
        self.assertAlmostEqual(1e-4, dsmax, 5)

        iforces, = input.IFORCE.get()
        self.assertEqual(2, len(iforces))

        kl, kc, kpar, icol, forcer, wlow, whig = iforces[0]
        self.assertEqual(1, kl)
        self.assertEqual(1, kc)
        self.assertEqual(1, kpar)
        self.assertEqual(4, icol)
        self.assertAlmostEqual(2000, forcer, 5)
        self.assertAlmostEqual(0.1, wlow, 5)
        self.assertAlmostEqual(2.0, whig, 5)

        kl, kc, kpar, icol, forcer, wlow, whig = iforces[1]
        self.assertEqual(1, kl)
        self.assertEqual(1, kc)
        self.assertEqual(1, kpar)
        self.assertEqual(5, icol)
        self.assertAlmostEqual(200, forcer, 5)
        self.assertAlmostEqual(0.1, wlow, 5)
        self.assertAlmostEqual(2.0, whig, 5)

        ibrspls, = input.IBRSPL.get()
        self.assertEqual(1, len(ibrspls))

        kl, kc, factor = ibrspls[0]
        self.assertEqual(1, kl)
        self.assertEqual(1, kc)
        self.assertAlmostEqual(2.0, factor, 5)

        ixrspls, = input.IXRSPL.get()
        self.assertEqual(1, len(ixrspls))

        kl, kc, factor = ibrspls[0]
        self.assertEqual(1, kl)
        self.assertEqual(1, kc)
        self.assertAlmostEqual(2.0, factor, 5)

        el, eu, nbe = input.NBE.get()
        self.assertAlmostEqual(0.0, el, 5)
        self.assertAlmostEqual(0.0, eu, 5)
        self.assertEqual(100, nbe)

        nbth, nbph = input.NBANGL.get()
        self.assertEqual(45, nbth)
        self.assertEqual(18, nbph)

        radm, nbre = input.EMERGP.get()
        self.assertAlmostEqual(0.005, radm, 5)
        self.assertEqual(100, nbre)

        energy_deposition_detectors, = input.energy_deposition_detectors.get()
        self.assertEqual(1, len(energy_deposition_detectors))

        el, eu, nbe, spectrum_filename, cylinders = energy_deposition_detectors[0]
        self.assertAlmostEqual(0.0, el, 5)
        self.assertAlmostEqual(0.0, eu, 5)
        self.assertEqual(100, nbe)
        self.assertIsNone(spectrum_filename)
        self.assertEqual(1, len(cylinders))

        kl, kc = cylinders[0]
        self.assertEqual(1, kl)
        self.assertEqual(1, kc)

        dose2d, = input.DOSE2D.get()
        self.assertEqual(1, len(dose2d))

        kl, kc, nz, nr = dose2d[0]
        self.assertEqual(1, kl)
        self.assertEqual(1, kc)
        self.assertEqual(nz, 100)
        self.assertEqual(nr, 50)

        filename, = input.RESUME.get()
        self.assertEqual('dump.dat', filename)

        filename, = input.DUMPTO.get()
        self.assertEqual('dump.dat', filename)

        dumpp, = input.DUMPP.get()
        self.assertAlmostEqual(60.0, dumpp, 5)

        seed1, seed2 = input.RSEED.get()
        self.assertEqual(1, seed1)
        self.assertEqual(1, seed2)

        dshn, = input.NSIMSH.get()
        self.assertAlmostEqual(1e9, dshn, 5)

        timea, = input.TIME.get()
        self.assertAlmostEqual(600.0, timea, 5)

    def test_example1_disc_write(self):
        input = create_example1_disc()
        input = self._write_read_input(input)
        self._test_example1_disc(input)

    def test_example1_disc_read(self):
        filepath = os.path.join(self.testdatadir, '1-disc', 'disc.in')
        input = PencylInput()
        with open(filepath, 'r') as fp:
            input.read(fp)
        self._test_example1_disc(input)

    def _test_example3_detector(self, input):
        geometry_definitions, = input.geometry_definitions.get()
        self.assertEqual(4, len(geometry_definitions))

        zlow, zhigh, xcen, ycen, cylinders = geometry_definitions[0]
        self.assertAlmostEqual(-0.24, zlow, 5)
        self.assertAlmostEqual(-0.16, zhigh, 5)
        self.assertAlmostEqual(0.0, xcen, 5)
        self.assertAlmostEqual(0.0, ycen, 5)
        self.assertEqual(1, len(cylinders))

        material, rin, rout = cylinders[0]
        self.assertEqual(3, material)
        self.assertAlmostEqual(0.0, rin, 5)
        self.assertAlmostEqual(4.05, rout, 5)

        zlow, zhigh, xcen, ycen, cylinders = geometry_definitions[1]
        self.assertAlmostEqual(-0.16, zlow, 5)
        self.assertAlmostEqual(0.0, zhigh, 5)
        self.assertIsNone(xcen)
        self.assertIsNone(ycen)
        self.assertEqual(2, len(cylinders))

        material, rin, rout = cylinders[0]
        self.assertEqual(2, material)
        self.assertAlmostEqual(0.0, rin, 5)
        self.assertAlmostEqual(3.97, rout, 5)

        material, rin, rout = cylinders[1]
        self.assertEqual(3, material)
        self.assertAlmostEqual(3.97, rin, 5)
        self.assertAlmostEqual(4.05, rout, 5)

        zlow, zhigh, xcen, ycen, cylinders = geometry_definitions[2]
        self.assertAlmostEqual(0.0, zlow, 5)
        self.assertAlmostEqual(7.72, zhigh, 5)
        self.assertIsNone(xcen)
        self.assertIsNone(ycen)
        self.assertEqual(3, len(cylinders))

        material, rin, rout = cylinders[0]
        self.assertEqual(1, material)
        self.assertAlmostEqual(0.0, rin, 5)
        self.assertAlmostEqual(3.81, rout, 5)

        material, rin, rout = cylinders[1]
        self.assertEqual(2, material)
        self.assertAlmostEqual(3.81, rin, 5)
        self.assertAlmostEqual(3.97, rout, 5)

        material, rin, rout = cylinders[2]
        self.assertEqual(3, material)
        self.assertAlmostEqual(3.97, rin, 5)
        self.assertAlmostEqual(4.05, rout, 5)

        zlow, zhigh, xcen, ycen, cylinders = geometry_definitions[3]
        self.assertAlmostEqual(7.72, zlow, 5)
        self.assertAlmostEqual(9.72, zhigh, 5)
        self.assertIsNone(xcen)
        self.assertIsNone(ycen)
        self.assertEqual(1, len(cylinders))

        material, rin, rout = cylinders[0]
        self.assertEqual(3, material)
        self.assertAlmostEqual(0.0, rin, 5)
        self.assertAlmostEqual(4.05, rout, 5)

        kparp, = input.SKPAR.get()
        self.assertEqual(2, kparp)

        se0, = input.SENERG.get()
        self.assertAlmostEqual(9.5e6, se0, 5)

        sx0, sy0, sz0 = input.SPOSIT.get()
        self.assertAlmostEqual(0.0, sx0, 5)
        self.assertAlmostEqual(0.0, sy0, 5)
        self.assertAlmostEqual(-10.0, sz0, 5)

        theta, phi, alpha = input.SCONE.get()
        self.assertAlmostEqual(0.0, theta, 5)
        self.assertAlmostEqual(0.0, phi, 5)
        self.assertAlmostEqual(0.0, alpha, 5)
#
        materials, = input.materials.get()
        self.assertEqual(3, len(materials))

        filename, eabs1, eabs2, eabs3, c1, c2, wcc, wcr = materials[0]
        self.assertEqual('NaI.mat', filename)
        self.assertAlmostEqual(5e4, eabs1, 5)
        self.assertAlmostEqual(5e3, eabs2, 5)
        self.assertAlmostEqual(5e4, eabs3, 5)
        self.assertAlmostEqual(0.1, c1, 5)
        self.assertAlmostEqual(0.1, c2, 5)
        self.assertAlmostEqual(2e3, wcc, 5)
        self.assertAlmostEqual(2e3, wcr, 5)

        filename, eabs1, eabs2, eabs3, c1, c2, wcc, wcr = materials[1]
        self.assertEqual('Al2O3.mat', filename)
        self.assertAlmostEqual(5e4, eabs1, 5)
        self.assertAlmostEqual(5e3, eabs2, 5)
        self.assertAlmostEqual(5e4, eabs3, 5)
        self.assertAlmostEqual(0.1, c1, 5)
        self.assertAlmostEqual(0.1, c2, 5)
        self.assertAlmostEqual(2e3, wcc, 5)
        self.assertAlmostEqual(2e3, wcr, 5)

        filename, eabs1, eabs2, eabs3, c1, c2, wcc, wcr = materials[2]
        self.assertEqual('Al.mat', filename)
        self.assertAlmostEqual(5e4, eabs1, 5)
        self.assertAlmostEqual(5e3, eabs2, 5)
        self.assertAlmostEqual(5e4, eabs3, 5)
        self.assertAlmostEqual(0.1, c1, 5)
        self.assertAlmostEqual(0.1, c2, 5)
        self.assertAlmostEqual(2e3, wcc, 5)
        self.assertAlmostEqual(2e3, wcr, 5)
#
        energy_deposition_detectors, = input.energy_deposition_detectors.get()
        self.assertEqual(1, len(energy_deposition_detectors))

        el, eu, nbe, spectrum_filename, cylinders = energy_deposition_detectors[0]
        self.assertAlmostEqual(0.0, el, 5)
        self.assertAlmostEqual(1e7, eu, 5)
        self.assertEqual(1000, nbe)
        self.assertIsNone(spectrum_filename)
        self.assertEqual(1, len(cylinders))

        kl, kc = cylinders[0]
        self.assertEqual(3, kl)
        self.assertEqual(1, kc)

        dose2d, = input.DOSE2D.get()
        self.assertEqual(1, len(dose2d))

        kl, kc, nz, nr = dose2d[0]
        self.assertEqual(3, kl)
        self.assertEqual(1, kc)
        self.assertEqual(nz, 50)
        self.assertEqual(nr, 50)

        filename, = input.RESUME.get()
        self.assertEqual('dump.dat', filename)

        filename, = input.DUMPTO.get()
        self.assertEqual('dump.dat', filename)

        dumpp, = input.DUMPP.get()
        self.assertAlmostEqual(60.0, dumpp, 5)

        dshn, = input.NSIMSH.get()
        self.assertAlmostEqual(1e8, dshn, 5)

        timea, = input.TIME.get()
        self.assertAlmostEqual(2e9, timea, 5)

    def test_example3_detector_write(self):
        input, _materials = create_example3_detector()
        input = self._write_read_input(input)
        self._test_example3_detector(input)

    def test_example3_detector_read(self):
        filepath = os.path.join(self.testdatadir, '3-detector', 'cyld.in')
        input = PencylInput()
        with open(filepath, 'r') as fp:
            input.read(fp)
        self._test_example3_detector(input)

if __name__ == '__main__': #pragma: no cover
    logging.basicConfig()
    logging.getLogger().setLevel(logging.DEBUG)
    unittest.main()



"""
Definition of base classes.
"""

# Standard library modules.
import abc
import re

# Third party modules.

# Local modules.

# Globals and constants variables.
LINE_KEYWORDS_SIZE = 6
LINE_SIZE = 80
SKIP_LINE = "       ."

PATTERN_LINE = re.compile(r'([A-Z0-9 ]{6})([\w\.\-\+ ]*)(\[.*\])?')

class _InputLineBase(metaclass=abc.ABCMeta):
    """
    Base class to parse and read PENELOPE text files.
    """

    def _parse_line(self, line):
        """
        Extracts the keyword, the values and the comment of an input line.
        The values are returned as a list.
        
        Args:
            line (str): Input line.

        Returns:
            tuple(str, tuple(str), str): Keyword, values, comment.
        """
        if line.startswith(' ' * 6):
            return None, None, line.strip()

        match = re.match(PATTERN_LINE, line)
        if not match:
            return None, None, None

        keyword, values, comment = match.groups()

        keyword = keyword.strip()
        values = tuple(values.split())
        if comment:
            comment = comment[1:-1]

        return keyword, values, comment

    def _create_line(self, name, values, comment=''):
        """
        Creates an input line of this keyword from the specified text.
        The white space between the items is automatically adjusted to fit the
        line size.
        The keyword and the total length of the line is checked not to exceed 
        their respective maximum size.
        
        Args:
            name (str): 6-character keyword.
            values (tuple or list): Values.
            comment (str, optional): Comment associated with the line.
            
        Returns:
            str: Formatted line.
        """
        # Keyword
        name = name.ljust(LINE_KEYWORDS_SIZE)
        assert len(name) == LINE_KEYWORDS_SIZE
        line = '{0} '.format(name.upper())

        # Values
        strvalues = []
        for value in values:
            if isinstance(value, float):
                strvalues.append('{0:g}'.format(value))
            else:
                strvalues.append(str(value))

        line += ' '.join(strvalues)

        # Comment
        if len(comment) > 0 and len(line) + len(comment) + 2 <= LINE_SIZE:
            line = line.ljust(LINE_SIZE - (len(comment) + 2))
            line += '[{0}]'.format(comment)

        if len(line) > LINE_SIZE:
            raise ValueError('Line of keyword {0} is too long, {1} > {2} characters'
                             .format(name, len(line), LINE_SIZE))

        return line

    def _peek_next_line(self, fileobj):
        """
        Returns the next line without advancing the current position.
        
        Args:
            fileobj (file object): File object opened with read access.
            
        Returns:
            str: Next line, stripped of all trailing white spaces.
        """
        # Remember the current position
        offset = fileobj.tell()

        # Read line
        line = self._read_next_line(fileobj)

        # Roll back to previous position
        fileobj.seek(offset)

        return line

    def _read_next_line(self, fileobj):
        """
        Returns the next line and advances the current position.
        Comment line (line starting with 7 spaces) are automatically skipped.
        
        Args:
            fileobj (file object): File object opened with read access.
            
        Returns:
            str: Next line, stripped of all trailing white spaces.
        """
        line = fileobj.readline().rstrip()

        # If line starts with 7 spaces, read next
        if line.startswith(' ' * 7):
            return self._read_next_line(fileobj)

        return line

    @abc.abstractmethod
    def read(self, fileobj):
        """
        Reads a PENELOPE-type file.
        
        Args:
            fileobj (file object): File object opened with read access.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def write(self, fileobj):
        """
        Writes to a PENELOPE-type file.
        
        Args:
            fileobj (file object): File object opened with write access.
        """
        raise NotImplementedError
